"""
Module with reading functionalities for Planck spectra.
"""

import os
import warnings

from configparser import ConfigParser
from typing import Dict, List, Optional, Tuple, Union

import numpy as np

from spectres.spectral_resampling_numba import spectres_numba
from typeguard import typechecked

from species.core import constants
from species.core.box import ColorMagBox, ColorColorBox, ModelBox, create_box
from species.phot.syn_phot import SyntheticPhotometry
from species.read.read_filter import ReadFilter
from species.util.dust_util import ism_extinction
from species.util.spec_util import create_wavelengths, smooth_spectrum


class ReadPlanck:
    """
    Class for reading a Planck spectrum.
    """

    @typechecked
    def __init__(
        self,
        wavel_range: Optional[
            Tuple[Union[float, np.float32], Union[float, np.float32]]
        ] = None,
        filter_name: Optional[str] = None,
    ) -> None:
        """
        Parameters
        ----------
        wavel_range : tuple(float, float), None
            Wavelength range (um). A wavelength range of 0.1-1000 um
            is used if set to ``None``. Not used if ``filter_name``
            is not set to ``None``.
        filter_name : str, None
            Filter name that is used for the wavelength range. The
            ``wavel_range`` is used if set to ``None``.

        Returns
        -------
        NoneType
            None
        """

        self.spectrum_interp = None
        self.wl_points = None
        self.wl_index = None

        self.filter_name = filter_name
        self.wavel_range = wavel_range

        if self.filter_name is not None:
            transmission = ReadFilter(self.filter_name)
            self.wavel_range = transmission.wavelength_range()

        elif self.wavel_range is None:
            self.wavel_range = (0.1, 1000.0)

        if 'SPECIES_CONFIG' in os.environ:
            config_file = os.environ['SPECIES_CONFIG']
        else:
            config_file = os.path.join(os.getcwd(), "species_config.ini")

        config = ConfigParser()
        config.read(config_file)

        self.database = config["species"]["database"]

    @staticmethod
    @typechecked
    def planck(
        wavel_points: np.ndarray, temperature: float, scaling: float
    ) -> np.ndarray:
        """
        Internal function for calculating a Planck function.

        Parameters
        ----------
        wavel_points : np.ndarray
            Wavelength points (um).
        temperature : float
            Temperature (K).
        scaling : float
            Scaling parameter.

        Returns
        -------
        np.ndarray
            Flux density (W m-2 um-1).
        """

        planck_1 = (
            2.0 * constants.PLANCK * constants.LIGHT**2 / (1e-6 * wavel_points) ** 5
        )

        planck_2 = (
            np.exp(
                constants.PLANCK
                * constants.LIGHT
                / (1e-6 * wavel_points * constants.BOLTZMANN * temperature)
            )
            - 1.0
        )

        return 1e-6 * np.pi * scaling * planck_1 / planck_2  # (W m-2 um-1)

    @staticmethod
    @typechecked
    def update_parameters(
        model_param: Dict[str, Union[float, List[float]]],
    ) -> Dict[str, float]:
        """
        Internal function for updating the dictionary with model
        parameters.

        Parameters
        ----------
        model_param : dict
            Dictionary with the 'teff' (K), 'radius' (Rjup), and
            'parallax' (mas) or 'distance' (pc). The values of
            'teff' and 'radius' can be a single float, or a list
            with floats for a combination of multiple Planck
            functions, e.g. {'teff': [1500., 1000.],
            'radius': [1., 2.], 'distance': 10.}.

        Returns
        -------
        dict
            Updated dictionary with model parameters.
        """

        updated_param = {}

        for i, _ in enumerate(model_param["teff"]):
            updated_param[f"teff_{i}"] = model_param["teff"][i]
            updated_param[f"radius_{i}"] = model_param["radius"][i]

        if "parallax" in model_param:
            updated_param["parallax"] = model_param["parallax"]
        elif "distance" in model_param:
            updated_param["distance"] = model_param["distance"]

        return updated_param

    @staticmethod
    @typechecked
    def apply_ext_ism(
        wavelengths: np.ndarray, flux: np.ndarray, v_band_ext: float, v_band_red: float
    ) -> np.ndarray:
        """
        Internal function for applying ISM extinction to a spectrum.

        wavelengths : np.ndarray
            Wavelengths (um) of the spectrum.
        flux : np.ndarray
            Fluxes (W m-2 um-1) of the spectrum.
        v_band_ext : float
            Extinction (mag) in the V band.
        v_band_red : float
            Reddening in the V band.

        Returns
        -------
        np.ndarray
            Fluxes (W m-2 um-1) with the extinction applied.
        """

        ext_mag = ism_extinction(v_band_ext, v_band_red, wavelengths)

        return flux * 10.0 ** (-0.4 * ext_mag)

    @typechecked
    def get_spectrum(
        self,
        model_param: Dict[str, Union[float, List[float]]],
        spec_res: Optional[float] = None,
        wavel_resample: Optional[np.ndarray] = None,
        **kwargs,
    ) -> ModelBox:
        """
        Function for calculating a Planck spectrum or a combination of
        multiple Planck spectra. The spectrum is calculated at
        :math:`R = 500`. Afterwards, an optional smoothing and
        wavelength resampling can be applied.

        Parameters
        ----------
        model_param : dict
            Dictionary with the 'teff' (K), 'radius' (Rjup), and
            'parallax' (mas) or 'distance' (pc). The values of
            'teff' and 'radius' can be a single float, or a list
            with floats for a combination of multiple Planck
            functions, e.g. {'teff': [1500., 1000.],
            'radius': [1., 2.], 'distance': 10.}.
        spec_res : float, None
            Spectral resolution that is used for smoothing the
            spectrum with a Gaussian kernel. No smoothing is
            applied if the argument is set to ``None``.
        wavel_resample : np.ndarray, None
            Wavelength points (:math:`\\mu`m) to which the spectrum
            will be resampled. The resampling is applied after the
            optional smoothing to the value of ``spec_res``.

        Returns
        -------
        species.core.box.ModelBox
            Box with the Planck spectrum.
        """

        if "smooth" in kwargs:
            warnings.warn(
                "The 'smooth' parameter has been "
                "deprecated. Please set only the "
                "'spec_res' argument, which can be set "
                "to None for not applying a smoothing.",
                DeprecationWarning,
            )

            if not kwargs["smooth"] and spec_res is not None:
                spec_res = None

        if "teff" in model_param and isinstance(model_param["teff"], list):
            model_param = self.update_parameters(model_param)

        wavel_points = create_wavelengths(self.wavel_range, 1000.0)

        n_planck = 0

        for item in model_param:
            if item[:4] == "teff":
                n_planck += 1

        if n_planck == 1:
            if "radius" in model_param and "parallax" in model_param:
                scaling = (
                    (model_param["radius"] * constants.R_JUP)
                    / (1e3 * constants.PARSEC / model_param["parallax"])
                ) ** 2

            elif "radius" in model_param and "distance" in model_param:
                scaling = (
                    (model_param["radius"] * constants.R_JUP)
                    / (model_param["distance"] * constants.PARSEC)
                ) ** 2

            else:
                scaling = 1.0

            flux = self.planck(
                wavel_points, model_param["teff"], scaling
            )  # (W m-2 um-1)

        else:
            flux = np.zeros(wavel_points.shape)

            for i in range(n_planck):
                if f"radius_{i}" in model_param and "parallax" in model_param:
                    scaling = (
                        (model_param[f"radius_{i}"] * constants.R_JUP)
                        / (1e3 * constants.PARSEC / model_param["parallax"])
                    ) ** 2

                elif f"radius_{i}" in model_param and "distance" in model_param:
                    scaling = (
                        (model_param[f"radius_{i}"] * constants.R_JUP)
                        / (model_param["distance"] * constants.PARSEC)
                    ) ** 2

                else:
                    scaling = 1.0

                flux += self.planck(
                    wavel_points, model_param[f"teff_{i}"], scaling
                )  # (W m-2 um-1)

        # Create ModelBox with the spectrum

        model_box = create_box(
            boxtype="model",
            model="planck",
            wavelength=wavel_points,
            flux=flux,
            parameters=model_param,
            quantity="flux",
        )

        # Apply extinction

        if "ism_ext" in model_param:
            ism_reddening = model_param.get("ism_red", 3.1)

            model_box.flux = self.apply_ext_ism(
                model_box.wavelength,
                model_box.flux,
                model_param["ism_ext"],
                ism_reddening,
            )

        # Smooth the spectrum

        if spec_res is not None:
            flux = smooth_spectrum(wavel_points, flux, spec_res)

        # Resample the spectrum

        if wavel_resample is not None:
            flux = spectres_numba(
                wavel_resample,
                wavel_points,
                flux,
                spec_errs=None,
                fill=np.nan,
                verbose=True,
            )

            model_box.wavelength = wavel_resample
            model_box.flux = flux

        lum_total = 0.0

        if n_planck == 1 and "radius" in model_param:
            lum_bb = (
                4.0
                * np.pi
                * (model_box.parameters["radius"] * constants.R_JUP) ** 2
                * constants.SIGMA_SB
                * model_box.parameters["teff"] ** 4.0
                / constants.L_SUN
            )  # (Lsun)

            lum_total += lum_bb
            model_box.parameters["log_lum"] = np.log10(lum_bb)

        elif n_planck > 1:
            for i in range(n_planck):
                if f"radius_{i}" in model_box.parameters:
                    # Add up the luminosity of the blackbody components (Lsun)
                    lum_bb = (
                        4.0
                        * np.pi
                        * (model_box.parameters[f"radius_{i}"] * constants.R_JUP) ** 2
                        * constants.SIGMA_SB
                        * model_box.parameters[f"teff_{i}"] ** 4.0
                        / constants.L_SUN
                    )

                    lum_total += lum_bb
                    model_box.parameters[f"luminosity_{i}"] = lum_bb

        if lum_total > 0.0:
            model_box.parameters["log_lum"] = np.log10(lum_total)

        return model_box

    @typechecked
    def get_flux(
        self, model_param: Dict[str, Union[float, List[float]]], synphot=None
    ) -> Tuple[float, None]:
        """
        Function for calculating the average flux
        density for the ``filter_name``.

        Parameters
        ----------
        model_param : dict
            Dictionary with the 'teff' (K), 'radius' (Rjup), and
            'parallax' (mas) or 'distance' (pc).
        synphot : SyntheticPhotometry, None
            Synthetic photometry object. The object is created if the
            argument is set to ``None``.

        Returns
        -------
        float
            Average flux density (W m-2 um-1).
        NoneType
            None
        """

        if "teff" in model_param and isinstance(model_param["teff"], list):
            model_param = self.update_parameters(model_param)

        spectrum = self.get_spectrum(model_param, 100.0)

        if synphot is None:
            synphot = SyntheticPhotometry(self.filter_name)

        return synphot.spectrum_to_flux(spectrum.wavelength, spectrum.flux)

    @typechecked
    def get_magnitude(
        self, model_param: Dict[str, Union[float, List[float]]], synphot=None
    ) -> Tuple[Tuple[float, None], Tuple[float, None]]:
        """
        Function for calculating the magnitude for the ``filter_name``.

        Parameters
        ----------
        model_param : dict
            Dictionary with the 'teff' (K), 'radius' (Rjup), and
            'parallax' (mas) or 'distance' (pc).
        synphot : SyntheticPhotometry, None
            Synthetic photometry object. The object is created if the
            argument is set to ``None``.

        Returns
        -------
        float
            Apparent magnitude (mag).
        float
            Absolute magnitude (mag)
        """

        if "teff" in model_param and isinstance(model_param["teff"], list):
            model_param = self.update_parameters(model_param)

        spectrum = self.get_spectrum(model_param, 100.0)

        if synphot is None:
            synphot = SyntheticPhotometry(self.filter_name)

        if "parallax" in model_param:
            distance = 1e3 / model_param["parallax"]
        else:
            distance = model_param["distance"]

        return synphot.spectrum_to_magnitude(
            spectrum.wavelength, spectrum.flux, distance=(distance, None)
        )

    @staticmethod
    @typechecked
    def get_color_magnitude(
        temperatures: np.ndarray,
        radius: float,
        filters_color: Tuple[str, str],
        filter_mag: str,
    ) -> ColorMagBox:
        """
        Function for calculating the colors and magnitudes in the range of 100-10000 K.

        Parameters
        ----------
        temperatures : np.ndarray
            Temperatures (K) for which the colors and magnitude are
            calculated.
        radius : float
            Radius (Rjup).
        filters_color : tuple(str, str)
            Filter names for the color.
        filter_mag : str
            Filter name for the absolute magnitudes.

        Returns
        -------
        species.core.box.ColorMagBox
            Box with the colors and magnitudes.
        """

        list_color = []
        list_mag = []

        for item in temperatures:
            model_param = {"teff": item, "radius": radius, "distance": 10.0}

            read_planck_0 = ReadPlanck(filter_name=filters_color[0])
            read_planck_1 = ReadPlanck(filter_name=filters_color[1])
            read_planck_2 = ReadPlanck(filter_name=filter_mag)

            app_mag_0, _ = read_planck_0.get_magnitude(model_param)
            app_mag_1, _ = read_planck_1.get_magnitude(model_param)
            app_mag_2, _ = read_planck_2.get_magnitude(model_param)

            list_color.append(app_mag_0[0] - app_mag_1[0])
            list_mag.append(app_mag_2[0])

        return create_box(
            boxtype="colormag",
            library="planck",
            object_type=None,
            filters_color=filters_color,
            filter_mag=filter_mag,
            color=list_color,
            magnitude=list_mag,
            sptype=temperatures,
            names=None,
        )

    @staticmethod
    @typechecked
    def get_color_color(
        temperatures: np.ndarray,
        radius: float,
        filters_colors: Tuple[Tuple[str, str], Tuple[str, str]],
    ) -> ColorColorBox:
        """
        Function for calculating two colors in the range of
        100-10000 K.

        Parameters
        ----------
        temperatures : np.ndarray
            Temperatures (K) for which the colors are calculated.
        radius : float
            Radius (Rjup).
        filters_colors : tuple(tuple(str, str), tuple(str, str))
            Two tuples with the filter names for the colors.

        Returns
        -------
        species.core.box.ColorColorBox
            Box with the colors.
        """

        list_color_1 = []
        list_color_2 = []

        for item in temperatures:
            model_param = {"teff": item, "radius": radius, "distance": 10.0}

            read_planck_0 = ReadPlanck(filter_name=filters_colors[0][0])
            read_planck_1 = ReadPlanck(filter_name=filters_colors[0][1])
            read_planck_2 = ReadPlanck(filter_name=filters_colors[1][0])
            read_planck_3 = ReadPlanck(filter_name=filters_colors[1][1])

            app_mag_0, _ = read_planck_0.get_magnitude(model_param)
            app_mag_1, _ = read_planck_1.get_magnitude(model_param)
            app_mag_2, _ = read_planck_2.get_magnitude(model_param)
            app_mag_3, _ = read_planck_3.get_magnitude(model_param)

            list_color_1.append(app_mag_0[0] - app_mag_1[0])
            list_color_2.append(app_mag_2[0] - app_mag_3[0])

        return create_box(
            boxtype="colorcolor",
            library="planck",
            object_type=None,
            filters=filters_colors,
            color1=list_color_1,
            color2=list_color_2,
            sptype=temperatures,
            names=None,
        )



"""
Module for isochrone data from evolutionary models.
"""

from typing import Optional

import h5py

from typeguard import typechecked

from species.data.isochrone_data.iso_ames import add_ames
from species.data.isochrone_data.iso_atmo import add_atmo
from species.data.isochrone_data.iso_baraffe2015 import add_baraffe2015
from species.data.isochrone_data.iso_btsettl import add_btsettl
from species.data.isochrone_data.iso_chabrier2023 import add_chabrier2023
from species.data.isochrone_data.iso_linder2019 import add_linder2019
from species.data.isochrone_data.iso_manual import add_manual
from species.data.isochrone_data.iso_marleau import add_marleau
from species.data.isochrone_data.iso_nextgen import add_nextgen
from species.data.isochrone_data.iso_parsec import add_parsec
from species.data.isochrone_data.iso_saumon2008 import add_saumon2008
from species.data.isochrone_data.iso_sonora_bobcat import add_sonora_bobcat
from species.data.isochrone_data.iso_sonora_diamondback import add_sonora_diamondback


@typechecked
def add_isochrone_grid(
    data_folder: str,
    hdf5_file: h5py._hl.files.File,
    model_name: Optional[str] = None,
    filename: Optional[str] = None,
    tag: Optional[str] = None,
) -> None:
    """
    Function for adding an isochrone grid to the database.

    Parameters
    ----------
    data_folder : str
        Folder where the data is located.
    hdf5_file : h5py._hl.files.File
        Database.
    model_name : str, None
        Evolutionary model ('ames', 'atmo', 'atmo-chabrier2023',
        'baraffe2015', 'bt-settl', 'linder2019', 'nextgen',
        'parsec', 'saumon2008', 'sonora-bobcat',
        'sonora-diamondback'). Isochrones will be automatically
        downloaded. Alternatively, the ``filename`` parameter
        can be used in combination with ``tag``.
    filename : str, None
        Filename with the isochrone data. The argument of
        ``model`` will be ignored by setting the argument
        of ``filename``. When using ``filename``, also
        the argument of ``tag`` should be set. Only files
        with isochrone data from
        https://phoenix.ens-lyon.fr/Grids/ and
        https://perso.ens-lyon.fr/isabelle.baraffe/ are
        supported. The parameter is ignored by setting
        the argument to ``None``.
    tag : str, None
        Database tag name where the isochrone that will be
        stored. Setting the argument is only required in
        combination with the ``filename`` parameter.
        Otherwise, the argument can be set to ``None``.

    Returns
    -------
    None
        NoneType
    """

    if model_name == "ames":
        add_ames(hdf5_file, data_folder)

    elif model_name == "atmo":
        add_atmo(hdf5_file, data_folder)

    elif model_name == "atmo-chabrier2023":
        add_chabrier2023(hdf5_file, data_folder)

    elif model_name == "baraffe2015":
        add_baraffe2015(hdf5_file, data_folder)

    elif model_name == "bt-settl":
        add_btsettl(hdf5_file, data_folder)

    elif model_name == "linder2019":
        add_linder2019(hdf5_file, data_folder)

    elif model_name == "marleau":
        add_marleau(hdf5_file, tag, filename)

    elif model_name == "nextgen":
        add_nextgen(hdf5_file, data_folder)

    elif model_name == "parsec":
        add_parsec(hdf5_file, data_folder)

    elif model_name == "saumon2008":
        add_saumon2008(hdf5_file, data_folder)

    elif model_name == "sonora-bobcat":
        add_sonora_bobcat(hdf5_file, data_folder)

    elif model_name == "sonora-diamondback":
        add_sonora_diamondback(hdf5_file, data_folder)

    else:
        add_manual(hdf5_file, tag, filename, model_name=tag)



"""
Utility functions for model spectra.
"""

import json
import warnings

from itertools import product
from pathlib import Path
from typing import Dict, Optional, Tuple, Union

import numpy as np

from astropy import units as u

# from PyAstronomy.pyasl import fastRotBroad
from scipy.interpolate import RegularGridInterpolator
from spectres.spectral_resampling_numba import spectres_numba
from typeguard import typechecked

from species.core import constants
from species.core.box import ModelBox, create_box
from species.util.dust_util import ism_extinction
from species.util.spec_util import create_wavelengths, smooth_spectrum


@typechecked
def convert_model_name(in_name: str) -> str:
    """
    Function for updating a model name for use in plots.

    Parameters
    ----------
    in_name : str
        Model name as used by species.

    Returns
    -------
    str
        Updated model name for plots.
    """

    data_file = (
        Path(__file__).parent.resolve().parents[0] / "data/model_data/model_data.json"
    )

    with open(data_file, "r", encoding="utf-8") as json_file:
        model_data = json.load(json_file)

    if in_name in model_data.keys():
        out_name = model_data[in_name]["name"]

    elif in_name == "planck":
        out_name = "Blackbody"

    else:
        out_name = in_name

        warnings.warn(
            f"The model name '{in_name}' is not known "
            "so the output name will not get adjusted "
            "for plot purposes"
        )

    return out_name


@typechecked
def powerlaw_spectrum(
    wavel_range: Union[Tuple[float, float], Tuple[np.float32, np.float32]],
    model_param: Dict[str, float],
    spec_res: float = 100.0,
) -> ModelBox:
    """
    Function for calculating a power-law spectrum. The power-law
    function is calculated in log(wavelength)-log(flux) space but
    stored in the :class:`~species.core.box.ModelBox` in linear
    wavelength-flux space.

    Parameters
    ----------
    wavel_range : tuple(float, float)
        Tuple with the minimum and maximum wavelength (um).
    model_param : dict
        Dictionary with the model parameters. Should contain
        `'log_powerlaw_a'`, `'log_powerlaw_b'`, and `'log_powerlaw_c'`.
    spec_res : float
        Spectral resolution (default: 100).

    Returns
    -------
    species.core.box.ModelBox
        Box with the power-law spectrum.
    """

    wavel = create_wavelengths((wavel_range[0], wavel_range[1]), spec_res)

    wavel *= 1e3  # (um) -> (nm)

    log_flux = (
        model_param["log_powerlaw_a"]
        + model_param["log_powerlaw_b"]
        * np.log10(wavel) ** model_param["log_powerlaw_c"]
    )

    model_box = create_box(
        boxtype="model",
        model="powerlaw",
        wavelength=1e-3 * wavel,  # (um)
        flux=10.0**log_flux,  # (W m-2 um-1)
        parameters=model_param,
        quantity="flux",
    )

    return model_box


@typechecked
def gaussian_spectrum(
    wavel_range: Union[Tuple[float, float], Tuple[np.float32, np.float32]],
    model_param: Dict[str, float],
    spec_res: float = 100.0,
    double_gaussian: bool = False,
) -> ModelBox:
    """
    Function for calculating a Gaussian spectrum (i.e. for an emission
    line).

    Parameters
    ----------
    wavel_range : tuple(float, float)
        Tuple with the minimum and maximum wavelength (um).
    model_param : dict
        Dictionary with the model parameters. Should contain
        ``'gauss_amplitude'``, ``'gauss_mean'``, ``'gauss_sigma'``,
        and optionally ``'gauss_offset'``.
    spec_res : float
        Spectral resolution (default: 100).
    double_gaussian : bool
        Set to ``True`` for returning a double Gaussian function.
        In that case, ``model_param`` should also contain
        ``'gauss_amplitude_2'``, ``'gauss_mean_2'``, and
        ``'gauss_sigma_2'``.

    Returns
    -------
    species.core.box.ModelBox
        Box with the Gaussian spectrum.
    """

    wavel = create_wavelengths((wavel_range[0], wavel_range[1]), spec_res)

    gauss_exp = np.exp(
        -0.5
        * (wavel - model_param["gauss_mean"]) ** 2
        / model_param["gauss_sigma"] ** 2
    )

    flux = model_param["gauss_amplitude"] * gauss_exp

    if double_gaussian:
        gauss_exp = np.exp(
            -0.5
            * (wavel - model_param["gauss_mean_2"]) ** 2
            / model_param["gauss_sigma_2"] ** 2
        )

        flux += model_param["gauss_amplitude_2"] * gauss_exp

    if "gauss_offset" in model_param:
        flux += model_param["gauss_offset"]

    model_box = create_box(
        boxtype="model",
        model="gaussian",
        wavelength=wavel,
        flux=flux,
        parameters=model_param,
        quantity="flux",
    )

    return model_box


@typechecked
def binary_to_single(param_dict: Dict[str, float], star_index: int) -> Dict[str, float]:
    """
    Function for converting a dictionary with atmospheric parameters
    of a binary system to a dictionary of parameters for one of the
    two stars.

    Parameters
    ----------
    param_dict : dict
        Dictionary with the atmospheric parameters of both stars. The
        keywords end either with ``_0`` or ``_1`` that correspond with
        ``star_index=0`` or ``star_index=1``.
    star_index : int
        Star index (0 or 1) that is used for the parameters in
        ``param_dict``.

    Returns
    -------
    dict
        Dictionary with the parameters of the selected star.
    """

    new_dict = {}

    for param_key, param_value in param_dict.items():
        if star_index == 0 and param_key[-1] == "0":
            new_dict[param_key[:-2]] = param_value

        elif star_index == 1 and param_key[-1] == "1":
            new_dict[param_key[:-2]] = param_value

        elif param_key[-2:] not in ["_0", "_1"]:
            new_dict[param_key] = param_value

    return new_dict


@typechecked
def extract_disk_param(
    param_dict: Dict[str, float], disk_index: Optional[int] = None
) -> Dict[str, float]:
    """
    Function for extracting the blackbody disk parameters from a
    dictionary with a mix of atmospheric and blackbody parameters.


    Parameters
    ----------
    param_dict : dict
        Dictionary with the model parameters.
    disk_index : int, None
        Disk index that is used for extracting the blackbody
        parameters from ``param_dict``. A single disk component
        will be used by setting the argument to ``None``.

    Returns
    -------
    dict
        Dictionary with the extracted blackbody parameters.
    """

    new_dict = {}

    if disk_index is None:
        new_dict["teff"] = param_dict["disk_teff"]
        new_dict["radius"] = param_dict["disk_radius"]

    else:
        new_dict["teff"] = param_dict[f"disk_teff_{disk_index}"]
        new_dict["radius"] = param_dict[f"disk_radius_{disk_index}"]

    for param_key, param_value in param_dict.items():
        if param_key in [
            "distance",
            "parallax",
            "ism_ext",
            "ext_av",
        ]:
            new_dict[param_key] = param_value

    return new_dict


@typechecked
def apply_obs(
    model_flux: np.ndarray,
    model_wavel: Optional[np.ndarray] = None,
    model_param: Optional[Dict[str, float]] = None,
    data_wavel: Optional[np.ndarray] = None,
    spec_res: Optional[float] = None,
    rot_broad: Optional[float] = None,
    rad_vel: Optional[float] = None,
    cross_sections: Optional[RegularGridInterpolator] = None,
    ext_model: Optional[str] = None,
) -> np.ndarray:
    """
    Function for post-processing of a model spectrum. This will
    apply a rotational broadening, radial velocity shift, extinction,
    scaling, instrumental broadening, and wavelength resampling.
    Each of the steps are optional, depending on the arguments that
    are set and the parameters in the ``model_param`` dictionary.

    Parameters
    ----------
    model_flux : np.ndarray
        Array with the fluxes of the model spectrum.
    model_wavel : np.ndarray, None
        Array with the wavelengths of the model spectrum. Used by
        most of the steps, except the flux scaling.
    model_param : dict(str, float), None
        Dictionary with the model parameters. Not used if the
        argument is set to ``None``.
    data_wavel : np.ndarray, None
        Array with the wavelengths of the data used for the
        wavelength resampling. Not applied if the argument is
        set to ``None``.
    spec_res : float, None
        Spectral resolution of the data used for the
        instrumental broadening. Not applied if the argument
        is set to ``None`` or np.nan.
    rot_broad : float, None
        Rotational broadening :math:`v\\sin{i}` (km/s). Not
        applied if the argument is set to ``None``.
    rad_vel : float, None
        Radial velocity (km/s). Not applied if the argument
        is set to ``None``.
    cross_sections : RegularGridInterpolator, None
        Interpolated cross sections for fitting extinction by dust
        grains with a log-normal or power-law size distribution.
    ext_model : str, None
        Name with the extinction model from the ``dust-extinction``
        package (see `list of available models
        <https://dust-extinction.readthedocs.io/en/latest/
        dust_extinction/choose_model.html>`_). For example,
        set the argument to ``'CCM89'`` to use the extinction
        relation from `Cardelli et al. (1989) <https://ui.adsabs.
        harvard.edu/abs/1989ApJ...345..245C/abstract>`_.

    Returns
    -------
    np.ndarray
        Array with the processed model fluxes.
    """

    # Set empty dictionary if None

    if model_param is None:
        model_param = {}

    # Apply rotational broadening

    if rot_broad is not None:
        model_flux = rot_int_cmj(
            wavel=model_wavel,
            flux=model_flux,
            vsini=rot_broad,
            eps=0.0,
        )

    # Apply extinction

    if "ism_ext" in model_param:
        ism_reddening = model_param.get("ism_red", 3.1)

        ext_filt = ism_extinction(
            model_param["ism_ext"],
            ism_reddening,
            model_wavel,
        )

        model_flux *= 10.0 ** (-0.4 * ext_filt)

    if "ext_av" in model_param:
        import dust_extinction.parameter_averages as dust_ext

        ext_object = getattr(dust_ext, ext_model)()

        if "ext_rv" in model_param:
            ext_object.Rv = model_param["ext_rv"]

        # Wavelength range (um) for which the extinction is defined
        ext_wavel = (1.0 / ext_object.x_range[1], 1.0 / ext_object.x_range[0])

        if model_wavel[0] < ext_wavel[0] or model_wavel[-1] > ext_wavel[1]:
            warnings.warn(
                "The wavelength range of the model spectrum "
                f"({model_wavel[0]:.3f}-{model_wavel[-1]:.3f} "
                "um) does not fully lie within the available "
                "wavelength range of the extinction model "
                f"({ext_wavel[0]:.3f}-{ext_wavel[1]:.3f} um). "
                "The extinction will therefore not be applied "
                "to fluxes of which the wavelength lies "
                "outside the range of the extinction model."
            )

        wavel_select = (model_wavel > ext_wavel[0]) & (model_wavel < ext_wavel[1])

        model_flux[wavel_select] *= ext_object.extinguish(
            model_wavel[wavel_select] * u.micron, Av=model_param["ext_av"]
        )

    elif "lognorm_ext" in model_param:
        cross_tmp = cross_sections(
            (
                model_wavel,
                10.0 ** model_param["lognorm_radius"],
                model_param["lognorm_sigma"],
            )
        )

        model_flux *= np.exp(-model_param["lognorm_ext"] * cross_tmp)

    elif "powerlaw_ext" in model_param:
        cross_tmp = cross_sections(
            (
                model_wavel,
                10.0 ** model_param["powerlaw_max"],
                model_param["powerlaw_exp"],
            )
        )

        model_flux *= np.exp(-model_param["powerlaw_ext"] * cross_tmp)

    # elif self.ext_filter is not None:
    #     ism_reddening = all_param.get("ism_red", 3.1)
    #
    #     av_required = convert_to_av(
    #         filter_name=self.ext_filter,
    #         filter_ext=all_param[f"phot_ext_{self.ext_filter}"],
    #         v_band_red=ism_reddening,
    #     )
    #
    #     ext_spec = ism_extinction(
    #         av_required, ism_reddening, self.spectrum[spec_item][0][:, 0]
    #     )
    #
    #     model_flux *= 10.0 ** (-0.4 * ext_spec)

    # Scale the spectrum by (radius/distance)^2

    if "radius" in model_param:
        model_flux *= (model_param["radius"] * constants.R_JUP) ** 2 / (
            1e3 * constants.PARSEC / model_param["parallax"]
        ) ** 2

    elif "flux_scaling" in model_param:
        model_flux *= model_param["flux_scaling"]

    elif "log_flux_scaling" in model_param:
        model_flux *= 10.0 ** model_param["log_flux_scaling"]

    # Apply radial velocity shift

    if rad_vel is not None:
        # Wavelength shift in um
        # rad_vel in km s-1 and constants.LIGHT in m s-1
        wavel_shift = model_wavel * 1e3 * rad_vel / constants.LIGHT

        # Resampling will introduce a few NaNs at the edge of the flux
        # array but that should not influence the fit given the total
        # number of wavelength points of a typical spectrum

        model_flux = spectres_numba(
            model_wavel,
            model_wavel + wavel_shift,
            model_flux,
            spec_errs=None,
            fill=np.nan,
            verbose=False,
        )

    # Apply instrument broadening

    if spec_res is not None and not np.isnan(spec_res):
        model_flux = smooth_spectrum(model_wavel, model_flux, spec_res)

    # Resample wavelengths to data

    if data_wavel is not None:
        # The 'fill' should not happen because model_wavel is
        # 20 wavelength points broader than data_wavel, but
        # spectres sometimes sets the outer fluxes to 'fill'
        # depending on the spectral resolution of the data

        model_flux = spectres_numba(
            data_wavel,
            model_wavel,
            model_flux,
            spec_errs=None,
            fill=np.nan,
            verbose=False,
        )

        # Set the flux to the neighboring wavelength
        # if a NaN is present at the edge.
        # TODO this is quick hack and not the best solution

        if np.isnan(model_flux[0]):
            model_flux[0] = model_flux[1]

        if np.isnan(model_flux[-1]):
            model_flux[0] = model_flux[-2]

    # Shift the spectrum by a constant

    # if "flux_offset" in model_param:
    #     model_flux += model_param["flux_offset"]

    return model_flux


@typechecked
def rot_int_cmj(
    wavel: np.ndarray,
    flux: np.ndarray,
    vsini: float,
    eps: float = 0.6,
    nr: int = 10,
    ntheta: int = 100,
    dif: float = 0.0,
):
    """
    A routine to quickly rotationally broaden a spectrum in linear time.
    This function has been adopted from `Carvalho & Johns-Krull (2023)
    <https://ui.adsabs.harvard.edu/abs/2023RNAAS...7...91C/abstract>`_.

    Parameters
    ----------
    wavel : np.ndarray
        Array with the wavelengths.
    flux : np.ndarray
        Array with the fluxes
    vsini : float
        Projected rotational velocity (km s-1).
    eps : float
        Coefficient of the limb darkening law (default: 0.0).
    nr : int
        Number of radial bins on the projected disk (default: 10).
    ntheta : int
        Number of azimuthal bins in the largest radial annulus
        (default: 100). Note: the number of bins at each r is
        int(r*ntheta) where r < 1.
    dif : float
        Differential rotation coefficient (default = 0.0), applied
        according to the law Omeg(th)/Omeg(eq) = (1 - dif/2 - (dif/2)
        cos(2 th)). Dif = 0.675 reproduces the law proposed by Smith,
        1994, A&A, Vol. 287, p. 523-534, to unify WTTS and CTTS.
        Dif = 0.23 is similar to observed solar differential rotation.
        Note: the th in the expression above is the stellar co-latitude,
        which is not the same as the integration variable used in the
        function. This is a disk integration routine.

    Returns
    -------
    np.ndarray
        Array with the rotationally broadened spectrum.
    """

    ns = np.zeros(flux.shape)
    tarea = 0.0
    dr = 1.0 / nr

    for j in range(nr):
        r = dr / 2.0 + j * dr
        area = (
            ((r + dr / 2.0) ** 2 - (r - dr / 2.0) ** 2)
            / int(ntheta * r)
            * (1.0 - eps + eps * np.cos(np.arcsin(r)))
        )

        for k in range(int(ntheta * r)):
            th = np.pi / int(ntheta * r) + k * 2.0 * np.pi / int(ntheta * r)

            if dif != 0:
                vl = (
                    vsini
                    * r
                    * np.sin(th)
                    * (
                        1.0
                        - dif / 2.0
                        - dif / 2.0 * np.cos(2.0 * np.arccos(r * np.cos(th)))
                    )
                )
            else:
                vl = r * vsini * np.sin(th)

            ns += area * np.interp(
                wavel + wavel * vl / (1e-3 * constants.LIGHT), wavel, flux
            )
            tarea += area

    return ns / tarea


@typechecked
def check_nearest_spec(model_name: str, model_param: Dict[str, float]):
    """
    Check if the nearest grid points of the requested model parameters
    have a spectrum stored in the database. For some grids, spectra
    are missing for certain parameters, in which case a spectrum with
    zero fluxes has been stored in the database. Interpolating from
    such a grid point will therefore give an inaccurate spectrum, so
    it is important to check if for example the best-fit parameters
    from a fit are close to grid points with a missing spectrum.

    Parameters
    ----------
    model_name : str
        Name of the atmosphere model.
    model_param : dict
        Dictionary with the model parameters.

    Returns
    -------
    NoneType
        None
    """

    from species.read.read_model import ReadModel

    read_model = ReadModel(model_name)
    model_points = read_model.get_points()

    near_low = {}
    near_high = {}
    param_idx = []
    for param_key, param_value in model_points.items():
        near_idx = np.argsort(np.abs(model_param[param_key] - param_value))
        near_low[param_key] = param_value[near_idx[0]]
        near_high[param_key] = param_value[near_idx[1]]
        param_idx.append((near_idx[0], near_idx[1]))

    for comb_item in list(product(*param_idx)):
        model_param = {}
        for param_idx, param_key in enumerate(model_points):
            model_param[param_key] = model_points[param_key][comb_item[param_idx]]

        model_box = read_model.get_data(model_param)

        if np.count_nonzero(model_box.flux) == 0:
            warnings.warn(
                "The selected parameters have a nearest grid point for "
                f"which a spectrum is not available: {model_param}. "
                "Therefore, zeros had been stored for the spectrum at "
                "this grid point. Interpolating from this grid point "
                "will therefore be inaccurate. When using FitModel, it "
                "is best to adjust the prior range in the 'bounds' "
                "parameter accordingly to exclude the parameter space "
                "for which model spectrum is missing. See also details "
                "printed when the model spectra are added to the "
                "database with add_model()."
            )



"""
Module with reading functionalities for calibration spectra.
"""

import configparser
import os

from typing import Dict, Optional, Tuple

import h5py
import numpy as np

from scipy.optimize import curve_fit
from spectres.spectral_resampling_numba import spectres_numba
from typeguard import typechecked

from species.core.box import SpectrumBox, create_box
from species.phot.syn_phot import SyntheticPhotometry
from species.read.read_filter import ReadFilter
from species.util.spec_util import create_wavelengths, smooth_spectrum


class ReadCalibration:
    """
    Class for reading a calibration spectrum from the database.
    """

    @typechecked
    def __init__(self, tag: str, filter_name: Optional[str] = None) -> None:
        """
        Parameters
        ----------
        tag : str
            Database tag of the calibration spectrum.
        filter_name : str, None
            Filter that is used for the wavelength range. The full
            spectrum is read if the argument is set to ``None``.

        Returns
        -------
        NoneType
            None
        """

        self.tag = tag
        self.filter_name = filter_name

        if filter_name is None:
            self.wavel_range = None

        else:
            transmission = ReadFilter(filter_name)
            self.wavel_range = transmission.wavelength_range()

        if 'SPECIES_CONFIG' in os.environ:
            config_file = os.environ['SPECIES_CONFIG']
        else:
            config_file = os.path.join(os.getcwd(), "species_config.ini")

        config = configparser.ConfigParser()
        config.read(config_file)

        self.database = config["species"]["database"]

    @typechecked
    def resample_spectrum(
        self,
        wavel_points: np.ndarray,
        model_param: Optional[Dict[str, float]] = None,
        spec_res: Optional[float] = None,
        apply_mask: bool = False,
        interp_highres: bool = False,
    ) -> SpectrumBox:
        """
        Function for resampling the spectrum and optional
        uncertainties onto a new wavelength grid.

        Parameters
        ----------
        wavel_points : np.ndarray
            Wavelengths (um).
        model_param : dict, None
            Dictionary with the model parameters, which should only
            contain the ``'scaling'`` keyword. No scaling is applied if
            the argument of ``model_param`` is set to ``None``.
        spec_res : float, None
            Spectral resolution that is used for smoothing the spectrum
            before resampling the wavelengths. No smoothing is applied
            if the argument is set to ``None``. The smoothing can only
            be applied to spectra with a constant
            $\\lambda/\\Delta\\lambda$ (which is the case for all model
            spectra that are supported by ``species``) or a constant
            wavelength spacing. The ``interp_highres`` parameter can be
            set to ``True`` to resample the spectrum to
            $\\lambda/\\Delta\\lambda = 10000$ before smoothing the
            spectrum.
        apply_mask : bool
            Exclude negative values and NaNs.
        interp_highres : bool
            Oversample the spectrum to
            $\\lambda/\\Delta\\lambda = 10000$, such that the smoothing
            by the ``spec_res`` parameter will be applied on a spectrum
            with constant logarithmically spaced wavelengths.

        Returns
        -------
        species.core.box.SpectrumBox
            Box with the resampled spectrum.
        """

        calib_box = self.get_spectrum(apply_mask=apply_mask)

        if interp_highres:
            wavel_highres = create_wavelengths(
                (calib_box.wavelength[0], calib_box.wavelength[-1]), 10000.0
            )

            flux_highres, error_highres = spectres_numba(
                wavel_highres,
                calib_box.wavelength,
                calib_box.flux,
                spec_errs=calib_box.error,
                fill=np.nan,
                verbose=True,
            )

            if spec_res is not None:
                flux_highres = smooth_spectrum(
                    wavelength=wavel_highres,
                    flux=flux_highres,
                    spec_res=spec_res,
                )

            flux_new, error_new = spectres_numba(
                wavel_points,
                wavel_highres,
                flux_highres,
                spec_errs=error_highres,
                fill=np.nan,
                verbose=True,
            )

        else:
            if spec_res is not None:
                calib_box.flux = smooth_spectrum(
                    wavelength=calib_box.wavelength,
                    flux=calib_box.flux,
                    spec_res=spec_res,
                )

            flux_new, error_new = spectres_numba(
                wavel_points,
                calib_box.wavelength,
                calib_box.flux,
                spec_errs=calib_box.error,
                fill=np.nan,
                verbose=True,
            )

        if model_param is not None:
            flux_new = model_param["scaling"] * flux_new
            error_new = model_param["scaling"] * error_new

        return create_box(
            boxtype="spectrum",
            spectrum="calibration",
            wavelength=wavel_points,
            flux=flux_new,
            error=error_new,
            name=self.tag,
        )

    @typechecked
    def get_spectrum(
        self,
        model_param: Optional[Dict[str, float]] = None,
        apply_mask: bool = False,
        wavel_sampling: Optional[float] = None,
        extrapolate: bool = False,
        min_wavelength: Optional[float] = None,
    ) -> SpectrumBox:
        """
        Function for selecting the calibration spectrum.

        Parameters
        ----------
        model_param : dict, None
            Model parameters. Should contain the 'scaling' value. Not
            used if set to ``None``.
        apply_mask : bool
            Exclude negative values and NaN values.
        wavel_sampling : float, None
            Wavelength sampling :math:`\\lambda/\\Delta\\lambda`. The
            original wavelength points are used if set to ``None``.
        extrapolate : bool
            Extrapolate to 6 um by fitting a power law function.
        min_wavelength : float, None
            Minimum wavelength used for fitting the power law function.
            All data is used if set to ``None``.

        Returns
        -------
        species.core.box.SpectrumBox
            Box with the spectrum.
        """

        with h5py.File(self.database, "r") as h5_file:
            cal_spec = np.array(h5_file[f"spectra/calibration/{self.tag}"])

            wavelength = cal_spec[0,]
            flux = cal_spec[1,]
            error = cal_spec[2,]

        if apply_mask:
            indices = np.where(flux > 0.0)[0]

            wavelength = wavelength[indices]
            flux = flux[indices]
            error = error[indices]

        if model_param is not None:
            flux = model_param["scaling"] * flux
            error = model_param["scaling"] * error

        if self.wavel_range is None:
            wl_index = np.ones(wavelength.size, dtype=bool)
        else:
            wl_index = (
                (flux > 0.0)
                & (wavelength > self.wavel_range[0])
                & (wavelength < self.wavel_range[1])
            )

        count = np.count_nonzero(wl_index)

        if count > 0:
            index = np.where(wl_index)[0]

            if index[0] > 0:
                wl_index[index[0] - 1] = True

            if index[-1] < len(wl_index) - 1:
                wl_index[index[-1] + 1] = True

            wavelength = wavelength[wl_index]
            flux = flux[wl_index]
            error = error[wl_index]

        if extrapolate:

            def _power_law(wavelength, offset, scaling, power_index):
                return offset + scaling * wavelength**power_index

            if min_wavelength:
                indices = np.where(wavelength > min_wavelength)[0]
            else:
                indices = np.arange(0, wavelength.size, 1)

            popt, pcov = curve_fit(
                f=_power_law,
                xdata=wavelength[indices],
                ydata=flux[indices],
                p0=(0.0, np.mean(flux[indices]), -1.0),
                sigma=error[indices],
            )

            sigma = np.sqrt(np.diag(pcov))

            print("Fit result for f(x) = a + b*x^c:")
            print(f"a = {popt[0]} +/- {sigma[0]}")
            print(f"b = {popt[1]} +/- {sigma[1]}")
            print(f"c = {popt[2]} +/- {sigma[2]}")

            while wavelength[-1] <= 6.0:
                wl_add = wavelength[-1] + wavelength[-1] / 1000.0

                wavelength = np.append(wavelength, wl_add)
                flux = np.append(flux, _power_law(wl_add, popt[0], popt[1], popt[2]))
                error = np.append(error, 0.0)

        if wavel_sampling is not None:
            wavelength_new = create_wavelengths(
                (wavelength[0], wavelength[-1]), wavel_sampling
            )

            flux_new, error_new = spectres_numba(
                wavelength_new,
                wavelength,
                flux,
                spec_errs=error,
                fill=np.nan,
                verbose=True,
            )

            # Fluxes at the edges on the resampled spectrum might be NaN

            flux_nan = np.isnan(flux_new)
            wavelength = wavelength_new[~flux_nan]
            flux = flux_new[~flux_nan]
            error = error_new[~flux_nan]

        return create_box(
            boxtype="spectrum",
            spectrum="calibration",
            wavelength=wavelength,
            flux=flux,
            error=error,
            name=self.tag,
        )

    @typechecked
    def get_flux(
        self, model_param: Optional[Dict[str, float]] = None
    ) -> Tuple[float, float]:
        """
        Function for calculating the average flux for the
        ``filter_name``.

        Parameters
        ----------
        model_param : dict, None
            Model parameters. Should contain the 'scaling' value. Not
            used if set to ``None``.

        Returns
        -------
        Returns
        -------
        float
            Average flux (W m-2 um-1).
        float
            Uncertainty (W m-2 um-1).
        """

        specbox = self.get_spectrum(model_param=model_param, apply_mask=True)

        synphot = SyntheticPhotometry(self.filter_name)

        return synphot.spectrum_to_flux(
            specbox.wavelength, specbox.flux, error=specbox.flux
        )

    @typechecked
    def get_magnitude(
        self,
        model_param: Optional[Dict[str, float]] = None,
        distance: Optional[Tuple[float, float]] = None,
    ) -> Tuple[Tuple[float, Optional[float]], Tuple[Optional[float], Optional[float]]]:
        """
        Function for calculating the apparent magnitude for the
        ``filter_name``.

        Parameters
        ----------
        model_param : dict, None
            Model parameters. Should contain the 'scaling' value. Not
            used if set to ``None``.
        distance : tuple(float, float), None
            Distance and uncertainty to the calibration object (pc).
            Not used if set to ``None``, in which case the returned
            absolute magnitude is ``(None, None)``.

        Returns
        -------
        tuple(float, float)
            Apparent magnitude and uncertainty.
        tuple(float, float), tuple(None, None)
            Absolute magnitude and uncertainty.
        """

        specbox = self.get_spectrum(model_param=model_param, apply_mask=True)

        if np.count_nonzero(specbox.error) == 0:
            error = None
        else:
            error = specbox.error

        synphot = SyntheticPhotometry(self.filter_name)

        return synphot.spectrum_to_magnitude(
            specbox.wavelength, specbox.flux, error=error, distance=distance
        )



import os
import shutil

import pytest
import numpy as np

from species import SpeciesInit
from species.data.database import Database
from species.read.read_planck import ReadPlanck
from species.phot.syn_phot import SyntheticPhotometry
from species.util import test_util


class TestPlanck:
    def setup_class(self):
        self.limit = 1e-8
        self.test_path = os.path.dirname(__file__) + "/"

    def teardown_class(self):
        os.remove("species_database.hdf5")
        os.remove("species_config.ini")
        shutil.rmtree("data/")

    def test_species_init(self):
        test_util.create_config("./")
        SpeciesInit()

    def test_read_planck(self):
        read_planck = ReadPlanck(filter_name="MKO/NSFCam.J")
        assert read_planck.wavel_range == pytest.approx(
            (1.1308, 1.3812), rel=1e-6, abs=0.0
        )

        read_planck = ReadPlanck(wavel_range=(1.0, 5.0))
        assert read_planck.wavel_range == (1.0, 5.0)

    def test_get_spectrum(self):
        read_planck = ReadPlanck(filter_name="MKO/NSFCam.J")
        modelbox = read_planck.get_spectrum(
            {"teff": 2000.0, "radius": 1.0, "parallax": 100.0}, 100.0
        )

        assert modelbox.model == "planck"
        assert modelbox.wavelength.shape == (204,)
        assert modelbox.flux.shape == (204,)

        assert np.sum(modelbox.wavelength) == pytest.approx(
            255.37728257033913, rel=self.limit, abs=0.0
        )
        assert np.sum(modelbox.flux) == pytest.approx(
            4.228433207324114e-12, rel=self.limit, abs=0.0
        )

    def test_get_flux(self):
        read_planck = ReadPlanck(filter_name="MKO/NSFCam.J")

        # low relative precision because of filter profile precision
        flux = read_planck.get_flux({"teff": 2000.0, "radius": 1.0, "distance": 10.0})
        assert flux[0] == pytest.approx(2.079882900702339e-14, rel=1e-4, abs=0.0)

        # low relative precision because of filter profile precision
        synphot = SyntheticPhotometry(filter_name="MKO/NSFCam.J")
        flux = read_planck.get_flux(
            {"teff": 2000.0, "radius": 1.0, "distance": 10.0}, synphot=synphot
        )
        assert flux[0] == pytest.approx(2.079882900702339e-14, rel=1e-4, abs=0.0)



import h5py
import numpy as np

from typeguard import typechecked

from species.core import constants


@typechecked
def add_manual(
    database: h5py._hl.files.File, tag: str, file_name: str, model_name: str
) -> None:
    """
    Function for adding any of the isochrones from
    https://phoenix.ens-lyon.fr/Grids/ or
    https://perso.ens-lyon.fr/isabelle.baraffe/ to
    the database.

    Parameters
    ----------
    database : h5py._hl.files.File
        Database.
    tag : str
        Tag name in the database.
    file_name : str
        Filename with the isochrones data.
    model_name : str
        Model name that is stored as attribute of the
        isochrone dataset in the HDF5 database.

    Returns
    -------
    NoneType
        None
    """

    # Read in all the data, ignoring empty lines or lines with "---"

    data = []

    check_baraffe = False
    baraffe_continue = False

    with open(file_name, encoding="utf-8") as open_file:
        for line in open_file:
            if "BHAC15" in line:
                check_baraffe = True
                continue

            if not baraffe_continue:
                if "(Gyr)" in line:
                    baraffe_continue = True
                else:
                    continue

            if line[0] == "!":
                line = line[1:]

            elif line[:2] == " !":
                line = line[2:]

            if "---" in line or line == "\n":
                continue

            data.append(list(filter(None, line.rstrip().split(" "))))

    iso_data = []

    for line in data:
        if "(Gyr)" in line:
            age = line[-1]

        elif "lg(g)" in line:
            # Isochrones from Phoenix website
            header = ["M/Ms", "Teff(K)"] + line[1:]

        elif "M/Ms" in line:
            # Isochrones from Baraffe et al. (2015)
            header = line.copy()

        else:
            line.insert(0, age)
            iso_data.append(line)

    header = np.asarray(header, dtype=str)
    iso_data = np.asarray(iso_data, dtype=float)

    iso_data[:, 0] *= 1e3  # (Myr)
    iso_data[:, 1] *= constants.M_SUN / constants.M_JUP  # (Mjup)

    if check_baraffe:
        iso_data[:, 5] *= constants.R_SUN / constants.R_JUP  # (Rjup)
    else:
        iso_data[:, 5] *= 1e9  # (cm)
        iso_data[:, 5] *= 1e-2 / constants.R_JUP  # (cm) -> (Rjup)

    index_sort = np.argsort(iso_data[:, 0])
    iso_data = iso_data[index_sort, :]

    print(f"\nAdding isochrones: {tag}...", end="", flush=True)

    if check_baraffe:
        filters = header[6:]
    else:
        filters = header[7:]

    dtype = h5py.string_dtype(encoding="utf-8", length=None)

    dset = database.create_dataset(
        f"isochrones/{tag}/filters", (np.size(filters),), dtype=dtype
    )

    dset[...] = filters

    dset = database.create_dataset(
        f"isochrones/{tag}/age", data=iso_data[:, 0]
    )  # (Myr)
    database.create_dataset(f"isochrones/{tag}/mass", data=iso_data[:, 1])  # (Mjup)
    database.create_dataset(f"isochrones/{tag}/teff", data=iso_data[:, 2])  # (K)
    database.create_dataset(
        f"isochrones/{tag}/log_lum", data=iso_data[:, 3]
    )  # log(L/Lsun)
    database.create_dataset(f"isochrones/{tag}/log_g", data=iso_data[:, 4])  # log(g)
    database.create_dataset(f"isochrones/{tag}/radius", data=iso_data[:, 5])  # (Rjup)

    if check_baraffe:
        database.create_dataset(f"isochrones/{tag}/lithium", data=iso_data[:, 6])
        database.create_dataset(f"isochrones/{tag}/magnitudes", data=iso_data[:, 7:])

    else:
        database.create_dataset(f"isochrones/{tag}/deuterium", data=iso_data[:, 6])
        database.create_dataset(f"isochrones/{tag}/lithium", data=iso_data[:, 7])
        database.create_dataset(f"isochrones/{tag}/magnitudes", data=iso_data[:, 8:])

    dset.attrs["model"] = model_name

    print(" [DONE]")
    print(f"Database tag: {tag}")



import os
import shutil

import pytest
import numpy as np

from species import SpeciesInit
from species.data.database import Database
from species.read.read_model import ReadModel
from species.util import test_util


class TestModel:
    def setup_class(self):
        self.limit = 1e-8
        self.test_path = os.path.dirname(__file__) + "/"
        self.model_param = {
            "teff": 2200.0,
            "logg": 4.5,
            "radius": 1.0,
            "parallax": 100.0,
        }

    def teardown_class(self):
        os.remove("species_database.hdf5")
        os.remove("species_config.ini")
        shutil.rmtree("data/")

    def test_species_init(self):
        test_util.create_config("./")
        SpeciesInit()

    def test_read_model(self):
        database = Database()

        database.add_model(
            "ames-cond",
            wavel_range=(1.0, 5.0),
            wavel_sampling=200.0,
            teff_range=(2000.0, 2500.0),
        )

        read_model = ReadModel("ames-cond")
        assert read_model.model == "ames-cond"

    def test_get_model(self):
        read_model = ReadModel("ames-cond", filter_name="Paranal/NACO.H")

        model_box = read_model.get_model(
            self.model_param,
            spec_res=100.0,
            magnitude=False,
        )

        assert np.sum(model_box.wavelength) == pytest.approx(
            152.40318409856752, rel=self.limit, abs=0.0
        )
        assert np.sum(model_box.flux) == pytest.approx(
            2.5045445767075435e-12, rel=self.limit, abs=0.0
        )

        model_box = read_model.get_model(
            self.model_param,
            spec_res=100.0,
            magnitude=True,
        )

        assert np.sum(model_box.wavelength) == pytest.approx(
            152.40318409856752, rel=self.limit, abs=0.0
        )
        assert np.sum(model_box.flux) == pytest.approx(
            1076.665025699217, rel=self.limit, abs=0.0
        )

    def test_get_data(self):
        read_model = ReadModel("ames-cond", filter_name="Paranal/NACO.H")
        model_box = read_model.get_data(self.model_param)

        assert np.sum(model_box.wavelength) == pytest.approx(
            152.40318409856752, rel=self.limit, abs=0.0
        )
        assert np.sum(model_box.flux) == pytest.approx(
            2.5044525130092417e-12, rel=self.limit, abs=0.0
        )

    def test_get_flux(self):
        read_model = ReadModel("ames-cond", filter_name="Paranal/NACO.H")
        flux = read_model.get_flux(self.model_param)

        assert flux[0] == pytest.approx(3.489600819641628e-14, rel=self.limit, abs=0.0)

    def test_get_magnitude(self):
        read_model = ReadModel("ames-cond", filter_name="Paranal/NACO.H")
        magnitude = read_model.get_magnitude(self.model_param)

        assert magnitude[0] == pytest.approx(
            11.291178975241221, rel=self.limit, abs=0.0
        )
        assert magnitude[1] == pytest.approx(
            11.291178975241221, rel=self.limit, abs=0.0
        )

    def test_get_bounds(self):
        read_model = ReadModel("ames-cond", filter_name="Paranal/NACO.H")
        bounds = read_model.get_bounds()

        assert bounds["teff"] == (2000.0, 2500.0)
        assert bounds["logg"] == (2.5, 5.5)

    def test_get_wavelengths(self):
        read_model = ReadModel("ames-cond", filter_name="Paranal/NACO.H")
        wavelengths = read_model.get_wavelengths()

        assert np.sum(wavelengths) == pytest.approx(
            813.2224003071026, rel=1e-7, abs=0.0
        )

    def test_get_points(self):
        read_model = ReadModel("ames-cond", filter_name="Paranal/NACO.H")
        points = read_model.get_points()

        assert np.sum(points["teff"]) == 13500.0
        assert np.sum(points["logg"]) == 28.0

    def test_get_parameters(self):
        read_model = ReadModel("ames-cond", filter_name="Paranal/NACO.H")
        parameters = read_model.get_parameters()

        assert parameters == ["teff", "logg"]



"""
Text
"""

import warnings

import numpy as np

from astroquery.simbad import Simbad
from astroquery.vizier import Vizier


# class NoStdStreams:
#     """
#     Text
#     """
#
#     def __init__(self, stdout=None, stderr=None):
#         self.devnull = open(os.devnull, "w", encoding="utf-8")
#         self._stdout = stdout or self.devnull or sys.stdout
#         self._stderr = stderr or self.devnull or sys.stderr
#         self.old_stdout = None
#         self.old_stderr = None
#
#     def __enter__(self):
#         self.old_stdout, self.old_stderr = sys.stdout, sys.stderr
#         self.old_stdout.flush()
#         self.old_stderr.flush()
#         sys.stdout, sys.stderr = self._stdout, self._stderr
#
#     def __exit__(self, exc_type, exc_value, traceback):
#         self._stdout.flush()
#         self._stderr.flush()
#         sys.stdout = self.old_stdout
#         sys.stderr = self.old_stderr
#         self.devnull.close()


# with NoStdStreams():
#     from astroquery.gaia import Gaia


def get_simbad(name):
    """
    Function for getting the SIMBAD identifier of an object.

    Parameters
    ----------
    name : np.ndarray

    Returns
    -------
    np.ndarray
        SIMBAD name.
    """

    simbad_result = Simbad.query_object(name)

    if simbad_result is None or len(simbad_result) == 0:
        simbad_id = None

    else:
        if "MAIN_ID" in simbad_result.columns:
            simbad_id = simbad_result["MAIN_ID"][0]
        else:
            simbad_id = simbad_result["main_id"][0]

    return simbad_id


def get_parallax(target):
    """
    Function for retrieving the parallax of an object.

    Parameters
    ----------
    target : str
        Target name.

    Returns
    -------
    str
        SIMBAD name.
    tuple(float, float)
        Parallax and uncertainty (mas).
    """

    # Liu et al. (2016)
    # http://cdsarc.u-strasbg.fr/viz-bin/cat/J/ApJ/833/96

    # Dupuy & Liu (2012)
    # http://cdsarc.u-strasbg.fr/viz-bin/cat/J/ApJS/201/19

    # Faherty et al. (2012)
    # http://cdsarc.u-strasbg.fr/viz-bin/cat/J/ApJ/752/56

    # Weinberger et al. (2016)
    # http://cdsarc.u-strasbg.fr/viz-bin/cat/J/AJ/152/24

    # Theissen (2018)
    # http://cdsarc.u-strasbg.fr/viz-bin/cat/J/ApJ/862/173

    # Kirkpatrick et al. (2012)
    # http://cdsarc.u-strasbg.fr/viz-bin/cat/J/ApJ/753/156

    # Monet et al. (1992)
    # http://cdsarc.u-strasbg.fr/viz-bin/cat/J/AJ/103/638

    # Faherty et al. (2009)
    # https://vizier.u-strasbg.fr/viz-bin/VizieR?-source=J/AJ/137/1

    catalogs = [
        "J/ApJ/833/96/sample",
        "J/ApJS/201/19/plx-phot",
        "J/ApJ/752/56/targets",
        "J/AJ/152/24",
        "J/ApJ/862/173/table",
        "J/AJ/103/638/table1",
        "J/AJ/103/638",
        "J/AJ/137/1/table4",
        "J/AJ/137/1",
    ]

    # if target[-2:] == 'AB':
    #     target = target[:-2]
    #
    # elif target[-3:] == 'ABC':
    #     target = target[:-3]
    #
    # if target[0:7] == 'DENIS-P':
    #     target = target[:5]+target[7:]
    #
    #     if target[-2] == '.':
    #         target = target[:-2]
    #
    # if target[0:5] == 'DENIS' and target[6:7] != 'J':
    #     target = target[:5]+' J'+target[6:]

    # time.sleep(0.15)

    simbad = Simbad()
    simbad.add_votable_fields("parallax")

    simbad_result = simbad.query_object(target)

    # query SIMBAD
    if simbad_result is not None and len(simbad_result) > 0:
        if "MAIN_ID" in simbad_result.columns:
            simbad_id = simbad_result["MAIN_ID"][0]
        else:
            simbad_id = simbad_result["main_id"][0]

        # For backward compatibility
        if not isinstance(simbad_id, str):
            simbad_id = simbad_id.decode("utf-8")

        if "PLX_VALUE" in simbad_result.columns:
            parallax = simbad_result["PLX_VALUE"][0]  # (mas)
        else:
            parallax = simbad_result["plx_value"][0]  # (mas)

        if "PLX_ERROR" in simbad_result.columns:
            parallax_error = simbad_result["PLX_ERROR"][0]  # (mas)
        else:
            parallax_error = simbad_result["plx_err"][0]  # (mas)

        if np.ma.is_masked(parallax):
            parallax = None

        if np.ma.is_masked(parallax_error):
            parallax_error = None

    else:
        simbad_id = None
        parallax = None
        parallax_error = None

    # Query VizieR catalogs
    if parallax is None:
        for item in catalogs:
            result = Vizier.query_object(target, catalog=item)

            if result.keys():
                if "plx" in result[0].keys():
                    parallax = result[0]["plx"][0]  # (mas)
                    if np.ma.is_masked(parallax):
                        parallax = None

                elif "Plx" in result[0].keys():
                    parallax = result[0]["Plx"][0]  # (mas)
                    if np.ma.is_masked(parallax):
                        parallax = None

                else:
                    parallax = None

                if "e_plx" in result[0].keys():
                    parallax_error = result[0]["e_plx"][0]  # (mas)
                    if np.ma.is_masked(parallax_error):
                        parallax_error = None

                elif "e_Plx" in result[0].keys():
                    parallax_error = result[0]["e_Plx"][0]  # (mas)
                    if np.ma.is_masked(parallax_error):
                        parallax_error = None

                else:
                    parallax_error = None

            if parallax is not None:
                break

    # query Gaia catalog
    # if np.ma.is_masked(parallax):
    #
    #     if simbad_result is not None:
    #         coord_ra = simbad_result["RA"][0]
    #         coord_dec = simbad_result["DEC"][0]
    #
    #         coord = SkyCoord(
    #             ra=coord_ra, dec=coord_dec, unit=(u.hourangle, u.deg), frame="icrs"
    #         )
    #
    #         result = Gaia.query_object(
    #             coordinate=coord, width=1.0 * u.arcsec, height=1.0 * u.arcsec
    #         )
    #
    #         if result:
    #             parallax = result["parallax"][0]  # (mas)

    if parallax is None:
        parallax = np.nan

    if parallax_error is None:
        parallax_error = np.nan

    if np.isnan(parallax):
        warnings.warn(
            f"No parallax was found for {target} " f"so storing a NaN value instead."
        )

    return simbad_id, (parallax, parallax_error)


# def get_parallax():
#     from species.data.database import Database
#     species_db = Database()
#     species_db.add_photometry("vlm-plx")
#
#     with h5py.File(species_db.database, "a") as hdf_file:
#         name = np.asarray(hdf_file["photometry/vlm-plx/name"])
#         ra_coord = np.asarray(hdf_file["photometry/vlm-plx/ra"])
#         dec_coord = np.asarray(hdf_file["photometry/vlm-plx/dec"])
#         distance = np.asarray(hdf_file["photometry/vlm-plx/distance"])
#         distance_error = np.asarray(hdf_file["photometry/vlm-plx/distance_error"])
#
#         simbad_id = []
#
#         print("Querying SIMBAD...", end="", flush=True)
#
#         for i, item in enumerate(name):
#             target_coord = SkyCoord(
#                 ra_coord[i], dec_coord[i], unit=(u.deg, u.deg), frame="icrs"
#             )
#
#             result_table = Simbad.query_region(target_coord, radius="0d0m2s")
#
#             if result_table is None:
#                 result_table = Simbad.query_region(target_coord, radius="0d0m5s")
#
#             if result_table is None:
#                 result_table = Simbad.query_region(target_coord, radius="0d0m20s")
#
#             if result_table is None:
#                 result_table = Simbad.query_region(target_coord, radius="0d1m0s")
#
#             if item == "HIP38939B":
#                 sim_id = get_simbad("HIP38939")
#             else:
#                 sim_id = result_table["MAIN_ID"][0]
#
#             # For backward compatibility
#             if not isinstance(sim_id, str):
#                 sim_id = sim_id.decode("utf-8")
#
#             simbad_id.append(sim_id)
#
#         print(" [DONE]")
#
#         simbad_id = np.asarray(simbad_id)
#
#         dtype = h5py.special_dtype(vlen=str)
#
#         dset = hdf_file.create_dataset(
#             "photometry/vlm-plx/simbad", (np.size(simbad_id),), dtype=dtype
#         )
#
#         dset[...] = simbad_id
#
#         np.savetxt(
#             "parallax.dat",
#             np.column_stack([name, simbad_id, distance, distance_error]),
#             header="VLM-PLX name - SIMBAD name - Distance (pc) - Error (pc)",
#             fmt="%35s, %35s, %8.2f, %8.2f",
#         )



import os
import shutil

import pytest
import numpy as np

from species import SpeciesInit
# from species.data.database import Database
# from species.read.read_spectrum import ReadSpectrum
# from species.plot.plot_spectrum import plot_spectrum
from species.util import test_util


class TestSpectrum:
    def setup_class(self):
        self.limit = 1e-8
        self.test_path = os.path.dirname(__file__) + "/"

    def teardown_class(self):
        os.remove("species_database.hdf5")
        os.remove("species_config.ini")
        # os.remove("spectrum.pdf")
        shutil.rmtree("data/")

    def test_species_init(self):
        test_util.create_config("./")
        SpeciesInit()

    # def test_read_spectrum(self):
    #     database = Database()
    #
    #     with pytest.warns(UserWarning):
    #         database.add_spectrum(
    #             "irtf",
    #             sptypes=[
    #                 "L",
    #             ],
    #         )
    #
    #     read_spectrum = ReadSpectrum("irtf", filter_name="MKO/NSFCam.H")
    #     assert read_spectrum.wavel_range == pytest.approx(
    #         (1.382, 1.8656), rel=1e-6, abs=0.0
    #     )
    #
    # def test_get_spectrum(self):
    #     read_spectrum = ReadSpectrum("irtf", filter_name="MKO/NSFCam.H")
    #     spec_box = read_spectrum.get_spectrum(
    #         sptypes=[
    #             "L0",
    #         ],
    #         exclude_nan=True,
    #     )
    #
    #     assert spec_box.wavelength[0].shape == (1063,)
    #     assert spec_box.flux[0].shape == (1063,)
    #
    #     assert np.sum(spec_box.wavelength[0]) == pytest.approx(
    #         1692.8604, rel=1e-7, abs=0.0
    #     )
    #     assert np.sum(spec_box.flux[0]) == pytest.approx(
    #         4.5681937e-11, rel=1e-7, abs=0.0
    #     )
    #
    #     plot_spectrum(
    #         boxes=[
    #             spec_box,
    #         ],
    #         filters=[
    #             "MKO/NSFCam.H",
    #         ],
    #         output="spectrum.pdf",
    #         xlim=(1.0, 2.5),
    #         offset=(-0.08, -0.06),
    #     )
    #
    #     assert os.path.exists("spectrum.pdf")



import os
import shutil

import pytest
import numpy as np

from species import SpeciesInit
from species.data.database import Database
from species.read.read_calibration import ReadCalibration
from species.util import test_util


class TestCalibration:
    def setup_class(self):
        self.limit = 1e-8
        self.test_path = os.path.dirname(__file__) + "/"
        self.model_param = {"scaling": 1.0}

    def teardown_class(self):
        os.remove("species_database.hdf5")
        os.remove("species_config.ini")
        shutil.rmtree("data/")

    def test_species_init(self):
        test_util.create_config("./")
        SpeciesInit()

    def test_read_calibration(self):
        database = Database()
        database.add_spectra("vega")

        read_calib = ReadCalibration("vega", filter_name="Paranal/NACO.H")
        assert read_calib.wavel_range == pytest.approx((1.44, 1.88), rel=1e-7, abs=0.0)

    def test_resample_spectrum(self):
        read_calib = ReadCalibration("vega")

        spec_box = read_calib.resample_spectrum(
            wavel_points=np.linspace(1.0, 5.0, 101),
            model_param=None,
            spec_res=None,
            apply_mask=True,
            interp_highres=False,
        )

        assert np.sum(spec_box.wavelength) == 303.0
        assert np.sum(spec_box.flux) == pytest.approx(
            6.51324713540645e-08, rel=self.limit, abs=0.0
        )

    def test_resample_smooth(self):
        read_calib = ReadCalibration("vega")

        spec_box = read_calib.resample_spectrum(
            wavel_points=np.linspace(1.0, 5.0, 101),
            model_param=None,
            spec_res=50.,
            apply_mask=True,
            interp_highres=True,
        )

        assert np.sum(spec_box.wavelength) == 303.0
        assert np.sum(spec_box.flux) == pytest.approx(
            6.516063242575011e-08, rel=self.limit, abs=0.0
        )

    def test_get_spectrum(self):
        read_calib = ReadCalibration("vega", filter_name="Paranal/NACO.Lp")
        spec_box = read_calib.get_spectrum(
            self.model_param, apply_mask=True, wavel_sampling=200.0
        )

        assert np.sum(spec_box.wavelength) == pytest.approx(
            175.68298611775307, rel=self.limit, abs=0.0
        )
        assert np.sum(spec_box.flux) == pytest.approx(
            2.3131999524734138e-09, rel=self.limit, abs=0.0
        )

    def test_get_flux(self):
        read_calib = ReadCalibration("vega", filter_name="Paranal/NACO.H")
        flux = read_calib.get_flux(model_param=self.model_param)

        assert flux[0] == pytest.approx(1.1149293297882683e-09, rel=self.limit, abs=0.0)

    def test_get_magnitude(self):
        read_calib = ReadCalibration("vega", filter_name="Paranal/NACO.H")
        app_mag, abs_mag = read_calib.get_magnitude(model_param=self.model_param)

        assert app_mag[0] == 0.03
        assert abs_mag[0] is None



"""
Module with reading functionalities for spectral libraries.
"""

import os

from configparser import ConfigParser
from typing import List, Optional, Union

import h5py
import numpy as np

from typeguard import typechecked

from species.core.box import PhotometryBox, SpectrumBox, create_box
from species.data.spec_data.add_spec_data import add_spec_library
from species.phot.syn_phot import SyntheticPhotometry
from species.read.read_filter import ReadFilter
from species.util.dust_util import ism_extinction


class ReadSpectrum:
    """
    Class for reading spectral library data from the database.
    """

    @typechecked
    def __init__(self, spec_library: str, filter_name: Optional[str] = None) -> None:
        """
        Parameters
        ----------
        spec_library : str
            Name of the spectral library ('irtf', 'spex', 'kesseli+2017',
            'bonnefoy+2014', 'allers+2013', or 'vega').
        filter_name : str, None
            Filter name for the wavelength range. Full spectra
            are read if the argument is set to ``None``.

        Returns
        -------
        NoneType
            None
        """

        self.spec_library = spec_library
        self.filter_name = filter_name

        if filter_name is None:
            self.wavel_range = None

        else:
            transmission = ReadFilter(filter_name)
            self.wavel_range = transmission.wavelength_range()

        if 'SPECIES_CONFIG' in os.environ:
            config_file = os.environ['SPECIES_CONFIG']
        else:
            config_file = os.path.join(os.getcwd(), "species_config.ini")

        config = ConfigParser()
        config.read(config_file)

        self.database = config["species"]["database"]
        self.data_folder = config["species"]["data_folder"]

    @typechecked
    def get_spectrum(
        self,
        object_name: Optional[Union[str, List[str]]] = None,
        sptypes: Optional[List[str]] = None,
        exclude_nan: bool = True,
        av_ext: Optional[float] = None,
    ) -> SpectrumBox:
        """
        Function for selecting spectra from the database.

        Parameters
        ----------
        object_name : str, list(str), None
            Object name of which the spectrum will be selected from the
            spectral library. Either a single object or a list of object
            names. By setting the argument to ``None``, the ``sptype``
            parameter will be used.
        sptypes : list(str), None
            Spectral types to select from a library. The spectral types
            should be indicated with two characters (e.g. 'M5', 'L2',
            'T3'). All spectra are selected if set to ``None``. For
            each object in the ``spec_library``, the requested
            ``sptypes`` are first compared with the optical spectral
            type and, if not available, secondly the near-infrared
            spectral type. To use the ``sptypes`` parameter, the
            argument of ``object_name`` should be set to ``None``
        exclude_nan : bool
            Exclude wavelength points for which the flux is NaN.
        av_ext : float, None
            Visual extinction, :math:`A_V`, applied to the spectra.
            The extinction is calculated with the empirical relation
            from `Cardelli et al. (1989)
            <https://ui.adsabs.harvard.edu/abs/1989ApJ...345..245C/
            abstract>`_. The extinction is not applied if the argument
            is set to ``None``.

        Returns
        -------
        species.core.box.SpectrumBox
            Box with the spectra.
        """

        with h5py.File(self.database, "r") as hdf5_file:
            # Check if the spectral library is found in
            # 'r' mode because the 'a' mode is not possible
            # when using multiprocessing
            spec_found = f"spectra/{self.spec_library}" in hdf5_file

        if not spec_found:
            with h5py.File(self.database, "a") as hdf5_file:
                add_spec_library(
                    self.data_folder, hdf5_file, self.spec_library, sptypes
                )

        list_wavelength = []
        list_flux = []
        list_error = []
        list_name = []
        list_simbad = []
        list_sptype = []
        list_parallax = []
        list_spec_res = []

        if object_name is not None:
            if isinstance(object_name, str):
                object_name = [object_name]

            with h5py.File(self.database, "r") as hdf5_file:
                for object_item in object_name:
                    if object_item not in hdf5_file[f"spectra/{self.spec_library}"]:
                        object_list = list(hdf5_file[f"spectra/{self.spec_library}"])

                        raise ValueError(
                            f"The selected 'object_name' (='{object_item}') "
                            "is found in the selected spectral library "
                            f"(='{self.spec_library}'). The following objects "
                            f"are available in the database: {object_list}"
                        )

        with h5py.File(self.database, "r") as hdf5_file:
            for spec_item in hdf5_file[f"spectra/{self.spec_library}"]:
                if object_name is not None and spec_item not in object_name:
                    continue

                dset = hdf5_file[f"spectra/{self.spec_library}/{spec_item}"]

                wavelength = dset[:, 0]  # (um)
                flux = dset[:, 1]  # (W m-2 um-1)
                error = dset[:, 2]  # (W m-2 um-1)

                if exclude_nan:
                    nan_index = np.isnan(flux)

                    wavelength = wavelength[~nan_index]
                    flux = flux[~nan_index]
                    error = error[~nan_index]

                if av_ext is not None:
                    ism_ext = ism_extinction(av_ext, 3.1, wavelength)
                    flux *= 10.0 ** (-0.4 * ism_ext)

                if self.wavel_range is None:
                    wl_index = np.arange(0, len(wavelength), 1)

                else:
                    wl_index = (
                        (flux > 0.0)
                        & (wavelength > self.wavel_range[0])
                        & (wavelength < self.wavel_range[1])
                    )

                count = np.count_nonzero(wl_index)

                if count > 0:
                    index = np.where(wl_index)[0]

                    if index[0] > 0:
                        wl_index[index[0] - 1] = True

                    if index[-1] < len(wl_index) - 1:
                        wl_index[index[-1] + 1] = True

                    list_wavelength.append(wavelength[wl_index])
                    list_flux.append(flux[wl_index])
                    list_error.append(error[wl_index])

                    attrs = dset.attrs

                    if "name" in attrs:
                        if isinstance(dset.attrs["name"], str):
                            list_name.append(dset.attrs["name"])
                        else:
                            list_name.append(dset.attrs["name"].decode("utf-8"))
                    else:
                        list_name.append("")

                    if "simbad" in attrs:
                        if isinstance(dset.attrs["simbad"], str):
                            list_simbad.append(dset.attrs["simbad"])
                        else:
                            list_simbad.append(dset.attrs["simbad"].decode("utf-8"))
                    else:
                        list_simbad.append("")

                    if "sptype" in attrs:
                        if isinstance(dset.attrs["sptype"], str):
                            list_sptype.append(dset.attrs["sptype"])
                        else:
                            list_sptype.append(dset.attrs["sptype"].decode("utf-8"))
                    else:
                        list_sptype.append("None")

                    if "parallax" in attrs:
                        list_parallax.append(
                            (dset.attrs["parallax"], dset.attrs["parallax_error"])
                        )
                    else:
                        list_parallax.append((np.nan, np.nan))

                    if "spec_res" in attrs:
                        list_spec_res.append(dset.attrs["spec_res"])
                    else:
                        list_spec_res.append(np.nan)

                else:
                    list_wavelength.append(np.array([]))
                    list_flux.append(np.array([]))
                    list_error.append(np.array([]))
                    list_name.append("")
                    list_simbad.append("")
                    list_sptype.append("None")
                    list_parallax.append((np.nan, np.nan))
                    list_spec_res.append(np.nan)

        spec_box = SpectrumBox()
        spec_box.spec_library = self.spec_library

        if sptypes is not None:
            spec_box.wavelength = []
            spec_box.flux = []
            spec_box.error = []
            spec_box.name = []
            spec_box.simbad = []
            spec_box.sptype = []
            spec_box.parallax = []
            spec_box.spec_res = []

            for spt_item in sptypes:
                for i, spec_item in enumerate(list_sptype):
                    if spt_item == spec_item[:2]:
                        spec_box.wavelength.append(list_wavelength[i])
                        spec_box.flux.append(list_flux[i])
                        spec_box.error.append(list_error[i])
                        spec_box.name.append(list_name[i])
                        spec_box.simbad.append(list_simbad[i])
                        spec_box.sptype.append(list_sptype[i])
                        spec_box.parallax.append(list_parallax[i])
                        spec_box.spec_res.append(list_spec_res[i])

        else:
            spec_box.wavelength = list_wavelength
            spec_box.flux = list_flux
            spec_box.error = list_error
            spec_box.name = list_name
            spec_box.simbad = list_simbad
            spec_box.sptype = list_sptype
            spec_box.parallax = list_parallax
            spec_box.spec_res = list_spec_res

        return spec_box

    @typechecked
    def get_flux(
        self,
        object_name: Optional[Union[str, List[str]]] = None,
        sptypes: Optional[List[str]] = None,
        av_ext: Optional[float] = None,
    ) -> PhotometryBox:
        """
        Function for calculating the flux density for the filter
        that is set with ``filter_name``.

        Parameters
        ----------
        object_name : str, list(str), None
            Object name of which the spectrum will be selected from the
            spectral library. Either a single object or a list of object
            names. By setting the argument to ``None``, the ``sptype``
            parameter will be used.
        sptypes : list(str), None
            Spectral types to select from a library. The spectral types
            should be indicated with two characters (e.g. 'M5', 'L2',
            'T3'). All spectra are selected if set to ``None``. For
            each object in the ``spec_library``, the requested
            ``sptypes`` are first compared with the optical spectral
            type and, if not available, secondly the near-infrared
            spectral type. To use the ``sptypes`` parameter, the
            argument of ``object_name`` should be set to ``None``
        av_ext : float, None
            Visual extinction, :math:`A_V`, applied to the spectra.
            The extinction is calculated with the empirical relation
            from `Cardelli et al. (1989)
            <https://ui.adsabs.harvard.edu/abs/1989ApJ...345..245C/
            abstract>`_. The extinction is not applied if the argument
            is set to ``None``.

        Returns
        -------
        species.core.box.PhotometryBox
            Box with the synthetic photometry.
        """

        spec_box = self.get_spectrum(
            object_name=object_name,
            sptypes=sptypes,
            exclude_nan=True,
            av_ext=av_ext,
        )

        n_spectra = len(spec_box.wavelength)

        filter_profile = ReadFilter(filter_name=self.filter_name)
        mean_wavel = filter_profile.mean_wavelength()

        filter_wavel_list = np.full(n_spectra, mean_wavel)
        filter_name_list = np.full(n_spectra, self.filter_name)

        syn_phot = SyntheticPhotometry(filter_name=self.filter_name)

        flux_list = []

        for i in range(n_spectra):
            spec_wavel = spec_box.wavelength[i]
            spec_flux = spec_box.flux[i]
            spec_error = spec_box.error[i]

            if np.sum(np.isnan(spec_error)) == spec_error.size:
                spec_error = None

            phot_flux = syn_phot.spectrum_to_flux(
                wavelength=spec_wavel,
                flux=spec_flux,
                error=spec_error,
            )

            flux_list.append(phot_flux)

        return create_box(
            boxtype="photometry",
            name=spec_box.name,
            sptype=spec_box.sptype,
            wavelength=filter_wavel_list,
            flux=np.array(flux_list),
            app_mag=None,
            abs_mag=None,
            filter_name=filter_name_list,
        )

    @typechecked
    def get_magnitude(
        self,
        object_name: Optional[Union[str, List[str]]] = None,
        sptypes: Optional[List[str]] = None,
        av_ext: Optional[float] = None,
    ) -> PhotometryBox:
        """
        Function for calculating the magnitude for the filter
        that is set with ``filter_name``.

        Parameters
        ----------
        object_name : str, list(str), None
            Object name of which the spectrum will be selected from the
            spectral library. Either a single object or a list of object
            names. By setting the argument to ``None``, the ``sptype``
            parameter will be used.
        sptypes : list(str), None
            Spectral types to select from a library. The spectral types
            should be indicated with two characters (e.g. 'M5', 'L2',
            'T3'). All spectra are selected if set to ``None``. For
            each object in the ``spec_library``, the requested
            ``sptypes`` are first compared with the optical spectral
            type and, if not available, secondly the near-infrared
            spectral type. To use the ``sptypes`` parameter, the
            argument of ``object_name`` should be set to ``None``
        av_ext : float, None
            Visual extinction, :math:`A_V`, applied to the spectra.
            The extinction is calculated with the empirical relation
            from `Cardelli et al. (1989)
            <https://ui.adsabs.harvard.edu/abs/1989ApJ...345..245C/
            abstract>`_. The extinction is not applied if the argument
            is set to ``None``.

        Returns
        -------
        species.core.box.PhotometryBox
            Box with the synthetic photometry.
        """

        spec_box = self.get_spectrum(
            object_name=object_name,
            sptypes=sptypes,
            exclude_nan=True,
            av_ext=av_ext,
        )

        n_spectra = len(spec_box.wavelength)

        filter_profile = ReadFilter(filter_name=self.filter_name)
        mean_wavel = filter_profile.mean_wavelength()

        filter_wavel_list = np.full(n_spectra, mean_wavel)
        filter_name_list = np.full(n_spectra, self.filter_name)

        syn_phot = SyntheticPhotometry(filter_name=self.filter_name)

        app_mag_list = []
        abs_mag_list = []

        for i in range(n_spectra):
            spec_wavel = spec_box.wavelength[i]
            spec_flux = spec_box.flux[i]
            spec_error = spec_box.error[i]

            if np.sum(np.isnan(spec_error)) == spec_error.size:
                spec_error = None

            if np.isnan(spec_box.parallax[i][0]):
                parallax = None

            else:
                parallax = (
                    float(spec_box.parallax[i][0]),
                    float(spec_box.parallax[i][1]),
                )

            app_mag, abs_mag = syn_phot.spectrum_to_magnitude(
                spec_wavel,
                spec_flux,
                error=spec_error,
                parallax=parallax,
            )

            app_mag_list.append(app_mag)
            abs_mag_list.append(abs_mag)

        return create_box(
            boxtype="photometry",
            name=spec_box.name,
            sptype=spec_box.sptype,
            wavelength=filter_wavel_list,
            flux=None,
            app_mag=np.asarray(app_mag_list),
            abs_mag=np.asarray(abs_mag_list),
            filter_name=filter_name_list,
        )



"""
Module for generating atmospheric model spectra with ``petitRADTRANS``.
Details on the radiative transfer, atmospheric setup, and opacities
can be found in `Mollière et al. (2019) <https://ui.adsabs.harvard.edu
/abs/2019A%26A...627A..67M/abstract>`_.
"""

import warnings

from typing import Dict, List, Optional, Tuple, Union

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import spectres

from matplotlib.ticker import MultipleLocator
from scipy.interpolate import interp1d
from spectres.spectral_resampling_numba import spectres_numba
from typeguard import typechecked

from species.core import constants
from species.core.box import ModelBox, create_box
from species.phot.syn_phot import SyntheticPhotometry
from species.read.read_filter import ReadFilter
from species.util.convert_util import logg_to_mass
from species.util.dust_util import apply_ism_ext
from species.util.model_util import rot_int_cmj
from species.util.retrieval_util import (
    calc_metal_ratio,
    calc_spectrum_clear,
    calc_spectrum_clouds,
    convolve_spectrum,
    log_x_cloud_base,
    pt_ret_model,
    pt_spline_interp,
    quench_pressure,
    scale_cloud_abund,
)


class ReadRadtrans:
    """
    Class for generating a model spectrum with ``petitRADTRANS``.
    """

    @typechecked
    def __init__(
        self,
        line_species: Optional[List[str]] = None,
        cloud_species: Optional[List[str]] = None,
        scattering: bool = False,
        wavel_range: Optional[Tuple[float, float]] = None,
        filter_name: Optional[str] = None,
        pressure_grid: str = "smaller",
        res_mode: str = "c-k",
        cloud_wavel: Optional[Tuple[float, float]] = None,
        max_press: Optional[float] = None,
        pt_manual: Optional[np.ndarray] = None,
        lbl_opacity_sampling: Optional[Union[int, np.int_]] = None,
    ) -> None:
        """
        Parameters
        ----------
        line_species : list, None
            List with the line species. No line species are used if set
            to ``None``.
        cloud_species : list, None
            List with the cloud species. No clouds are used if set to
            ``None``.
        scattering : bool
            Include scattering in the radiative transfer.
        wavel_range : tuple(float, float), None
            Wavelength range (:math:`\\mu`m). The wavelength range is
            set to 0.8-10.0 :math:`\\mu`m if set to ``None`` or not
            used if ``filter_name`` is not ``None``.
        filter_name : str, None
            Filter name that is used for the wavelength range. The
            ``wavel_range`` is used if ``filter_name`` is set to
            ``None``.
        pressure_grid : str
            The type of pressure grid that is used for the radiative
            transfer. Either 'standard', to use 180 layers both for
            the atmospheric structure (e.g. when interpolating the
            abundances) and 180 layers with the radiative transfer,
            or 'smaller' to use 60 (instead of 180) with the radiative
            transfer, or 'clouds' to start with 1440 layers but
            resample to ~100 layers (depending on the number of cloud
            species) with a refinement around the cloud decks. For
            cloudless atmospheres it is recommended to use 'smaller',
            which runs faster than 'standard' and provides sufficient
            accuracy. For cloudy atmosphere, one can test with
            'smaller' but it is recommended to use 'clouds' for
            improved accuracy fluxes.
        res_mode : str
            Resolution mode ('c-k' or 'lbl'). The low-resolution mode
            ('c-k') calculates the spectrum with the correlated-k
            assumption at :math:`\\lambda/\\Delta \\lambda = 1000`. The
            high-resolution mode ('lbl') calculates the spectrum with a
            line-by-line treatment at
            :math:`\\lambda/\\Delta \\lambda = 10^6`.
        cloud_wavel : tuple(float, float), None
            Tuple with the wavelength range (:math:`\\mu`m) that is
            used for calculating the median optical depth of the
            clouds at the gas-only photosphere and then scaling the
            cloud optical depth to the value of ``log_tau_cloud``.
            The range of ``cloud_wavel`` should be encompassed by
            the range of ``wavel_range``.  The full wavelength
            range (i.e. ``wavel_range``) is used if the argument is
            set to ``None``.
        max_pressure : float, None
            Maximum pressure (bar) for the free temperature nodes.
            The default value is set to 1000 bar.
        pt_manual : np.ndarray, None
            A 2D array that contains the P-T profile that is used
            when ``pressure_grid="manual"``. The shape of array should
            be (n_pressure, 2), with pressure (bar) as first column
            and temperature (K) as second column. It is recommended
            that the pressures are logarithmically spaced.
        lbl_opacity_sampling : int, None
            This is the same parameter as in ``petitRADTRANS`` which is
            used with ``res_mode='lbl'`` to downsample the line-by-line
            opacities by selecting every ``lbl_opacity_sampling``-th
            wavelength from the original sampling of
            :math:`\\lambda/\\Delta \\lambda = 10^6`. Setting this
            parameter will lower the computation time. By setting the
            argument to ``None``, the original sampling is used so no
            downsampling is applied.

        Returns
        -------
        NoneType
            None
        """

        # Set several of the required ReadRadtrans attributes

        self.filter_name = filter_name
        self.wavel_range = wavel_range
        self.scattering = scattering
        self.pressure_grid = pressure_grid
        self.cloud_wavel = cloud_wavel
        self.pt_manual = pt_manual
        self.res_mode = res_mode

        if lbl_opacity_sampling is None:
            self.lbl_opacity_sampling = 1
        else:
            self.lbl_opacity_sampling = lbl_opacity_sampling

        # Set maximum pressure

        if max_press is None:
            self.max_press = 1e3
        else:
            self.max_press = max_press

        # Set the wavelength range

        if self.filter_name is not None:
            transmission = ReadFilter(self.filter_name)
            self.wavel_range = transmission.wavelength_range()
            self.wavel_range = (0.9 * self.wavel_range[0], 1.2 * self.wavel_range[1])

        elif self.wavel_range is None:
            self.wavel_range = (0.8, 10.0)

        # Set the list with line species

        if line_species is None:
            self.line_species = []
        else:
            self.line_species = line_species

        # Set the list with cloud species and the number of P-T points

        if cloud_species is None:
            self.cloud_species = []
        else:
            self.cloud_species = cloud_species

        # Set the number of pressures

        if self.pressure_grid in ["standard", "smaller"]:
            # Initiate 180 pressure layers but use only
            # 60 layers during the radiative transfer
            # when pressure_grid is set to 'smaller'
            n_pressure = 180

        elif self.pressure_grid == "clouds":
            # Initiate 1140 pressure layers but use fewer
            # layers (~100) during the radiative tranfer
            # after running make_half_pressure_better
            n_pressure = 1440

        else:
            raise ValueError(
                f"The argument of pressure_grid "
                f"('{self.pressure_grid}') is "
                f"not recognized. Please use "
                f"'standard', 'smaller', or 'clouds'."
            )

        # Create 180 pressure layers in log space

        if self.pressure_grid == "manual":
            if self.pt_manual is None:
                raise UserWarning(
                    "A 2D array with the P-T profile "
                    "should be provided as argument "
                    "of pt_manual when using "
                    "pressure_grid='manual'."
                )

            self.pressure = self.pt_manual[:, 0]

        else:
            self.pressure = np.logspace(-6, np.log10(self.max_press), n_pressure)

        # Import petitRADTRANS here because it is slow

        print("Importing petitRADTRANS...", end="", flush=True)

        from petitRADTRANS.radtrans import Radtrans
        from petitRADTRANS.chemistry.utils import simplify_species_list
        from petitRADTRANS.chemistry.pre_calculated_chemistry import (
            PreCalculatedEquilibriumChemistryTable,
        )

        print(" [DONE]")

        # Pressure array for Radtrans

        if self.pressure_grid in ["standard", "manual"]:
            radtrans_press = np.copy(self.pressure)

        elif self.pressure_grid == "smaller":
            radtrans_press = self.pressure[::3]

        elif self.pressure_grid == "clouds":
            radtrans_press = self.pressure[::24]

        # Create the Radtrans object

        self.rt_object = Radtrans(
            pressures=radtrans_press,
            line_species=self.line_species,
            rayleigh_species=["H2", "He"],
            cloud_species=self.cloud_species,
            gas_continuum_contributors=["H2-H2", "H2-He"],
            wavelength_boundaries=self.wavel_range,
            line_opacity_mode=self.res_mode,
            scattering_in_emission=self.scattering,
            line_by_line_opacity_sampling=self.lbl_opacity_sampling,
        )

        # Initialize attributes to None

        self.abund_smooth = None

        # Load equilibrium chemistry table

        self.eq_chem = PreCalculatedEquilibriumChemistryTable()
        self.eq_chem.load()

    @typechecked
    def get_model(
        self,
        model_param: Dict[str, float],
        quenching: Optional[str] = None,
        spec_res: Optional[float] = None,
        wavel_resample: Optional[np.ndarray] = None,
        plot_contribution: Optional[Union[bool, str]] = False,
        temp_nodes: Optional[Union[int, np.integer]] = None,
    ) -> ModelBox:
        """
        Function for calculating a model spectrum with
        radiative transfer code of ``petitRADTRANS``.

        Parameters
        ----------
        model_param : dict
            Dictionary with the model parameters. Various
            parameterizations can be used for the
            pressure-temperature (P-T) profile, abundances
            (chemical equilibrium or free abundances), and
            the cloud properties. The type of parameterizations
            that will be used depend on the parameters provided
            in the dictionary of ``model_param``. Below is an
            (incomplete) list of the supported parameters.

            Mandatory parameters:

                - The surface gravity, ``logg``, should always
                  be included. It is provided in cgs units as
                  :math:`\\log_{10}{g}`.

            Scaling parameters (optional):

                - The radius (:math:`R_\\mathrm{J}`), ``radius``,
                  and parallax (mas), ``parallax``, are optional
                  parameters that can be included for scaling the
                  flux from the planet surface to the observer.

                - Instead of ``parallax``, it is also possible to
                  provided the distance (pc) with the ``distance``
                  parameter.

            Chemical abundances (mandatory -- one of the options
            should be used):

                - Chemical equilibrium requires the ``metallicity``,
                  ``c_o_ratio`` parameters. Optionally, the
                  ``log_p_quench`` (as :math:`\\log_{10}P/\\mathrm{bar}`)
                  can be included for setting a quench pressure for
                  CO/CH$_4$/H$_2$O. If this last parameter is used,
                  then the argument of ``quenching`` should be set
                  to ``'pressure'``.

                - Free abundances requires the parameters that have the
                  names from ``line_species`` and ``cloud_species``.
                  These will be used as :math:`\\log_{10}` mass fraction
                  of the line and cloud species. For example, if
                  ``line_species`` includes ``H2O_HITEMP``
                  then ``model_param`` should contain the ``H2O_HITEMP``
                  parameter. For a mass fraction of :math:`10^{-3}` the
                  dictionary value can be set to -3. Or, if
                  ``cloud_species`` contains ``MgSiO3(c)_cd`` then
                  ``model_param`` should contain the ``MgSiO3(c)``
                  parameter. So it is provided without the suffix,
                  ``_cd``, for the particle shape and structure.

            Pressure-temperature (P-T) profiles (mandatory -- one of
            the options should be used):

                - Eddington approximation requires the ``tint``
                  and ``log_delta`` parameters.

                - Parametrization from `Mollière et al (2020)
                  <https://ui.adsabs.harvard.edu/abs/2020A%26A...640A.
                  131M/abstract>`_ that was used for HR 8799 e. It
                  requires ``tint``, ``alpa``, ``log_delta``, ``t1``,
                  ``t2``, and ``t3`` as parameters.

                - Arbitrary number of free temperature nodes requires
                  parameters ``t0``, ``t1``, ``t2``, etc. So counting
                  from zero up to the number of nodes that are
                  required. The nodes will be interpolated to a larger
                  number of points in log-pressure space (set with the
                  ``pressure_grid`` parameter) by using a cubic spline.
                  Optionally, the ``pt_smooth`` parameter can also be
                  included in ``model_param``, which is used for
                  smoothing the interpolated P-T profile with a
                  Gaussian kernel in :math:`\\log{P/\\mathrm{bar}}`.
                  A recommended value for the kernel is 0.3 dex,
                  so ``pt_smooth=0.3``.

                - Instead of a parametrization, it is also possible
                  to provide a manual P-T profile as ``numpy`` array
                  with the argument of ``pt_manual``.

            Cloud models (optional -- one of the options can be used):

                - Physical clouds as in `Mollière et al (2020)
                  <https://ui.adsabs.harvard.edu/abs/2020A%26A...640A.
                  131M/abstract>`_ require the parameters ``fsed``,
                  ``log_kzz``, and ``sigma_lnorm``. Cloud abundances
                  are either specified relative to the equilibrium
                  abundances (when using chemical equilibrium
                  abundances for the line species) or as free
                  abundances (when using free abundances for the line
                  species). For the first case, the relative mass
                  fractions are specified for example with the
                  ``mgsio3_fraction`` parameter if the list with
                  ``cloud_species`` contains ``MgSiO3(c)_cd``.

                - With the physical clouds, instead of including the
                  mass fraction with the ``_fraction`` parameters,
                  it is also possible to enforce the clouds (to ensure
                  an effect on the spectrum) by scaling the opacities
                  with the ``log_tau_cloud`` parameter. This is the
                  wavelength-averaged optical depth of the clouds down
                  to the gas-only photosphere. The abundances are
                  now specified relative to the first cloud species
                  that is listed in ``cloud_species``. The ratio
                  parameters should be provided with the ``_ratio``
                  suffix. For example, if
                  ``cloud_species=['MgSiO3(c)_cd', 'Fe(c)_cd',
                  'Al2O3(c)_cd']`` then the ``fe_mgsio3_ratio`` and
                  ``al2o3_mgsio3_ratio`` parameters are required.

                - Instead of a single sedimentation parameter,
                  ``fsed``, it is also possible to include two values,
                  ``fsed_1`` and ``fsed_2``. This will calculate a
                  weighted combination of two cloudy spectra, to mimic
                  horizontal cloud variations. The weight should be
                  provided with the ``f_clouds`` parameter (between
                  0 and 1) in the ``model_param`` dictionary.

                - Parametrized cloud opacities with a cloud absorption
                  opacity, ``log_kappa_abs``, and powerlaw index,
                  ``opa_abs_index``. Furthermore, ``log_p_base`` and
                  ``fsed`` are required parameters. In addition to
                  absorption, parametrized scattering opacities are
                  added with the optional ``log_kappa_sca`` and
                  ``opa_sca_index`` parameters. Optionally, the
                  ``lambda_ray`` can be included, which is the
                  wavelength at which the opacity changes to a
                  :math:`\\lambda^{-4}` dependence in the Rayleigh
                  regime. It is also possible to include
                  ``log_tau_cloud``, which can be used for
                  enforcing clouds in the photospheric region by
                  scaling the cloud opacities.

                - Parametrized cloud opacities with a total cloud
                  opacity, ``log_kappa_0``, and a single scattering
                  albedo, ``albedo``. Furthermore, ``opa_index``,
                  ``log_p_base``, and ``fsed``, are required
                  parameters. This is `cloud model 2` from
                  `Mollière et al (2020) <https://ui.adsabs.harvard.
                  edu/abs/2020A%26A...640A.131M/abstract>`_
                  Optionally, ``log_tau_cloud`` can be used for
                  enforcing clouds in the photospheric region by
                  scaling the cloud opacities.

                - Gray clouds are simply parametrized with the
                  ``log_kappa_gray`` and ``log_cloud_top``
                  parameters. These clouds extend from the bottom
                  of the atmosphere up to the cloud top pressure and
                  have a constant opacity. Optionally, a single
                  scattering albedo, ``albedo``, can be specified.
                  Also ``log_tau_cloud`` can be used for enforcing
                  clouds in the photospheric region by scaling the
                  cloud opacities.

            Extinction (optional):

                 - Extinction can optionally be applied to the spectrum
                   by including the ``ism_ext`` parameter, which is the
                   the visual extinction, $A_V$. The empirical relation
                   from `Cardelli et al. (1989) <https://ui.adsabs.
                   harvard.edu/abs/1989ApJ...345..245C/abstract>`_
                   is used for calculating the extinction at other
                   wavelengths.

                 - When using ``ism_ext``, the reddening, $R_V$, can
                   also be optionally set with the ``ism_red``
                   parameter. Otherwise it is set to the standard
                   value for the diffuse ISM, $R_V = 3.1$.

            Radial velocity and broadening (optional):

                 - Radial velocity shift can be applied by adding the
                   ``rad_vel`` parameter. This shifts the spectrum
                   by a constant velocity (km/s).

                 - Rotational broadening can be applied by adding the
                   ``vsini`` parameter, which is the projected spin
                   velocity (km/s), :math:`v\\sin{i}`. The broadening
                   is applied with the function from `Carvalho &
                   Johns-Krull (2023) <https://ui.adsabs.harvard.edu/
                   abs/2023RNAAS...7...91C/abstract>`_.

        quenching : str, None
            Quenching type for CO/CH$_4$/H$_2$O abundances. Either
            the quenching pressure (bar) is a free parameter
            (``quenching='pressure'``) or the quenching pressure is
            calculated from the mixing and chemical timescales
            (``quenching='diffusion'``). The quenching is not applied
            if the argument is set to ``None``.
        spec_res : float, None
            Spectral resolution, achieved by smoothing with a Gaussian
            kernel. No smoothing is applied when the argument is set
            to ``None``.
        wavel_resample : np.ndarray, None
            Wavelength points (:math:`\\mu`m) to which the spectrum
            will be resampled. The original wavelengths points will
            be used if the argument is set to ``None``.
        plot_contribution : bool, str, None
            Filename for the plot with the emission contribution. The
            plot is not created if the argument is set to ``False`` or
            ``None``. If set to ``True``, the plot is shown in an
            interface window instead of written to a file.
        temp_nodes : int, None
            Number of free temperature nodes.

        Returns
        -------
        species.core.box.ModelBox
            Box with the petitRADTRANS model spectrum.
        """

        # Set chemistry type

        if "metallicity" in model_param and "c_o_ratio" in model_param:
            chemistry = "equilibrium"

        else:
            chemistry = "free"

            check_nodes = {}

            for line_item in self.line_species:
                abund_count = 0

                for node_idx in range(100):
                    if f"{line_item}_{node_idx}" in model_param:
                        abund_count += 1
                    else:
                        break

                check_nodes[line_item] = abund_count

            # Check if there are an equal number of
            # abundance nodes for all the line species

            nodes_list = list(check_nodes.values())

            if not all(value == nodes_list[0] for value in nodes_list):
                raise ValueError(
                    "The number of abundance nodes is "
                    "not equal for all the lines "
                    f"species: {check_nodes}"
                )

            if all(value == 0 for value in nodes_list):
                abund_nodes = None
            else:
                abund_nodes = nodes_list[0]

            for line_item in self.line_species:
                if abund_nodes is None:
                    if line_item not in model_param:
                        raise RuntimeError(
                            f"The abundance of {line_item} is not "
                            "found in the dictionary with parameters "
                            "of 'model_param'. Please add the log10 "
                            f"mass fraction of {line_item}."
                        )

                else:
                    for node_idx in range(abund_nodes):
                        if f"{line_item}_{node_idx}" not in model_param:
                            raise RuntimeError(
                                f"The abundance of {line_item} is not "
                                "found in the dictionary with parameters "
                                "of 'model_param'. Please add the log10 "
                                f"mass fraction of {line_item}."
                            )

        # Check quenching parameter

        if not hasattr(self, "quenching"):
            self.quenching = quenching

        if self.quenching is not None and chemistry != "equilibrium":
            raise ValueError(
                "The 'quenching' parameter can only be used in combination with "
                "chemistry='equilibrium'."
            )

        if self.quenching is not None and self.quenching not in [
            "pressure",
            "diffusion",
        ]:
            raise ValueError(
                "The argument of 'quenching' should be of the following: "
                "'pressure', 'diffusion', or None."
            )

        # Abundance nodes

        if chemistry == "free" and abund_nodes is not None:
            knot_press_abund = np.logspace(
                np.log10(self.pressure[0]), np.log10(self.pressure[-1]), abund_nodes
            )

        else:
            knot_press_abund = None

        # C/O and [Fe/H]

        if chemistry == "equilibrium":
            # Equilibrium chemistry
            metallicity = model_param["metallicity"]
            c_o_ratio = model_param["c_o_ratio"]

            log_x_abund = None

        elif chemistry == "free":
            # Free chemistry

            # TODO Set [Fe/H] = 0 for Molliere P-T profile and
            # cloud condensation profiles
            metallicity = 0.0

            # Get smoothing parameter for abundance profiles

            if "abund_smooth" in model_param:
                self.abund_smooth = model_param["abund_smooth"]

            # Create a dictionary with the mass fractions

            if abund_nodes is None:
                log_x_abund = {}
                for line_item in self.line_species:
                    log_x_abund[line_item] = model_param[line_item]

                _, _, c_o_ratio = calc_metal_ratio(log_x_abund, self.line_species)

            else:
                log_x_abund = {}
                for line_item in self.line_species:
                    for node_idx in range(abund_nodes):
                        log_x_abund[f"{line_item}_{node_idx}"] = model_param[
                            f"{line_item}_{node_idx}"
                        ]

                # TODO Set C/O = 0.55 for Molliere P-T profile
                # and cloud condensation profiles
                c_o_ratio = 0.55

        # Create the P-T profile

        if self.pressure_grid == "manual":
            temp = self.pt_manual[:, 1]

        elif (
            "tint" in model_param
            and "log_delta" in model_param
            and "alpha" in model_param
        ):
            temp, _, _ = pt_ret_model(
                np.array([model_param["t1"], model_param["t2"], model_param["t3"]]),
                10.0 ** model_param["log_delta"],
                model_param["alpha"],
                model_param["tint"],
                self.pressure,
                metallicity,
                c_o_ratio,
            )

        elif "tint" in model_param and "log_delta" in model_param:
            tau = self.pressure * 1e6 * 10.0 ** model_param["log_delta"]
            temp = (0.75 * model_param["tint"] ** 4.0 * (2.0 / 3.0 + tau)) ** 0.25

        elif "PTslope_1" in model_param:
            num_layer = 6  # could make a variable in the future
            layer_pt_slopes = np.ones(num_layer) * np.nan
            for index in range(num_layer):
                layer_pt_slopes[index] = model_param[f"PTslope_{num_layer - index}"]

            try:
                from petitRADTRANS.physics import dTdP_temperature_profile
            except ImportError as import_error:
                raise ImportError(
                    "Can't import the dTdP profile function from "
                    "petitRADTRANS, check that your version of pRT "
                    "includes this function in petitRADTRANS.physics",
                    import_error,
                )

            temp = dTdP_temperature_profile(
                self.pressure,
                num_layer,  # could change in the future
                layer_pt_slopes,
                model_param["T_bottom"],
            )

        else:
            if temp_nodes is None:
                temp_nodes = 0

                for temp_idx in range(100):
                    if f"t{temp_idx}" in model_param:
                        temp_nodes += 1
                    else:
                        break

            knot_press = np.logspace(
                np.log10(self.pressure[0]), np.log10(self.pressure[-1]), temp_nodes
            )

            knot_temp = []
            for temp_idx in range(temp_nodes):
                knot_temp.append(model_param[f"t{temp_idx}"])

            knot_temp = np.asarray(knot_temp)

            if "pt_smooth" in model_param:
                pt_smooth = model_param["pt_smooth"]

            else:
                pt_smooth = None

            temp = pt_spline_interp(
                knot_press,
                knot_temp,
                self.pressure,
                pt_smooth=pt_smooth,
            )

        # Set the log quenching pressure, log(P/bar)

        if self.quenching == "pressure":
            p_quench = 10.0 ** model_param["log_p_quench"]

        elif self.quenching == "diffusion":
            p_quench = quench_pressure(
                self.pressure,
                temp,
                model_param["metallicity"],
                model_param["c_o_ratio"],
                model_param["logg"],
                model_param["log_kzz"],
            )

        else:
            if "log_p_quench" in model_param:
                warnings.warn(
                    "The 'model_param' dictionary contains the "
                    "'log_p_quench' parameter but 'quenching=None'. "
                    "The quenching pressure from the dictionary is "
                    "therefore ignored."
                )

            p_quench = None

        if (
            len(self.cloud_species) > 0
            or "log_kappa_0" in model_param
            or "log_kappa_gray" in model_param
            or "log_kappa_abs" in model_param
        ):
            tau_cloud = None
            log_x_base = None

            if (
                "log_kappa_0" in model_param
                or "log_kappa_gray" in model_param
                or "log_kappa_abs" in model_param
            ):
                if "log_tau_cloud" in model_param:
                    tau_cloud = 10.0 ** model_param["log_tau_cloud"]

                elif "tau_cloud" in model_param:
                    tau_cloud = model_param["tau_cloud"]

            elif chemistry == "equilibrium":
                # Create the dictionary with the mass fractions of the
                # clouds relative to the maximum values allowed from
                # elemental abundances

                cloud_fractions = {}

                for item in self.cloud_species:
                    if f"{item}_fraction" in model_param:
                        cloud_fractions[item] = model_param[f"{item}_fraction"]

                    elif f"{item}_tau" in model_param:
                        # Import the chemistry module here in case
                        # petitRADTRANS was not installed

                        # Interpolate the abundances, following chemical equilibrium

                        abund_in, mmw, _ = self.eq_chem.interpolate_mass_fractions(
                            np.full(self.pressure.size, c_o_ratio),
                            np.full(self.pressure.size, metallicity),
                            temp,
                            self.pressure,
                            carbon_pressure_quench=p_quench,
                            full=True,
                        )

                        # Calculate the scaled mass fraction of the clouds

                        cloud_fractions[item] = scale_cloud_abund(
                            model_param,
                            self.rt_object,
                            self.pressure,
                            temp,
                            mmw,
                            "equilibrium",
                            abund_in,
                            item,
                            model_param[f"{item}_tau"],
                            pressure_grid=self.pressure_grid,
                        )

                if "log_tau_cloud" in model_param:
                    # Set the log mass fraction to zero and use the
                    # optical depth parameter to scale the cloud mass
                    # fraction with petitRADTRANS

                    tau_cloud = 10.0 ** model_param["log_tau_cloud"]

                elif "tau_cloud" in model_param:
                    # Set the log mass fraction to zero and use the
                    # optical depth parameter to scale the cloud mass
                    # fraction with petitRADTRANS

                    tau_cloud = model_param["tau_cloud"]

                if tau_cloud is not None:
                    for i, item in enumerate(self.cloud_species):
                        if i == 0:
                            cloud_fractions[item] = 0.0

                        else:
                            cloud_fractions[item] = model_param[
                                f"{item}_{self.cloud_species[0]}_ratio"
                            ]

                # Create a dictionary with the log mass fractions at the cloud base

                log_x_base = log_x_cloud_base(c_o_ratio, metallicity, cloud_fractions)

            elif chemistry == "free":
                # Add the log10 mass fractions of the clouds to the dictionary

                log_x_base = {}

                if "log_tau_cloud" in model_param:
                    # Set the log mass fraction to zero and use the
                    # optical depth parameter to scale the cloud mass
                    # fraction with petitRADTRANS

                    tau_cloud = 10.0 ** model_param["log_tau_cloud"]

                elif "tau_cloud" in model_param:
                    # Set the log mass fraction to zero and use the
                    # optical depth parameter to scale the cloud mass
                    # fraction with petitRADTRANS

                    tau_cloud = model_param["tau_cloud"]

                if tau_cloud is None:
                    for item in self.cloud_species:
                        # Set the log10 of the mass fractions at th
                        # cloud base equal to the value from the
                        # parameter dictionary
                        log_x_base[item] = model_param[item]

                else:
                    # Set the log10 of the mass fractions with the
                    # ratios from the parameter dictionary and
                    # scale to the actual mass fractions with
                    # tau_cloud that is used in calc_spectrum_clouds
                    for i, item in enumerate(self.cloud_species):
                        if i == 0:
                            log_x_base[item] = 0.0

                        else:
                            log_x_base[item] = model_param[
                                f"{item}_{self.cloud_species[0]}_ratio"
                            ]

            # Calculate the petitRADTRANS spectrum
            # for a cloudy atmosphere

            if "fsed_1" in model_param and "fsed_2" in model_param:
                cloud_dict = model_param.copy()
                cloud_dict["fsed"] = cloud_dict["fsed_1"]

                (
                    wavelength,
                    flux_1,
                    extra_out_1,
                    _,
                ) = calc_spectrum_clouds(
                    self.rt_object,
                    self.pressure,
                    temp,
                    c_o_ratio,
                    metallicity,
                    p_quench,
                    log_x_abund,
                    log_x_base,
                    cloud_dict,
                    model_param["logg"],
                    knot_press_abund=knot_press_abund,
                    abund_smooth=self.abund_smooth,
                    pressure_grid=self.pressure_grid,
                    plotting=False,
                    contribution=True,
                    tau_cloud=tau_cloud,
                    cloud_wavel=self.cloud_wavel,
                    return_opacities=True,
                    eq_chem=self.eq_chem,
                )

                cloud_dict = model_param.copy()
                cloud_dict["fsed"] = cloud_dict["fsed_2"]

                (
                    wavelength,
                    flux_2,
                    extra_out_2,
                    _,
                ) = calc_spectrum_clouds(
                    self.rt_object,
                    self.pressure,
                    temp,
                    c_o_ratio,
                    metallicity,
                    p_quench,
                    log_x_abund,
                    log_x_base,
                    cloud_dict,
                    model_param["logg"],
                    knot_press_abund=knot_press_abund,
                    abund_smooth=self.abund_smooth,
                    pressure_grid=self.pressure_grid,
                    plotting=False,
                    contribution=True,
                    tau_cloud=tau_cloud,
                    cloud_wavel=self.cloud_wavel,
                    return_opacities=True,
                    eq_chem=self.eq_chem,
                )

                flux = (
                    model_param["f_clouds"] * flux_1
                    + (1.0 - model_param["f_clouds"]) * flux_2
                )

                emission_contr = (
                    model_param["f_clouds"] * extra_out_1["emission_contribution"]
                    + (1.0 - model_param["f_clouds"])
                    * extra_out_2["emission_contribution"]
                )

            else:
                (
                    wavelength,
                    flux,
                    extra_out,
                    _,
                ) = calc_spectrum_clouds(
                    self.rt_object,
                    self.pressure,
                    temp,
                    c_o_ratio,
                    metallicity,
                    p_quench,
                    log_x_abund,
                    log_x_base,
                    model_param,
                    model_param["logg"],
                    knot_press_abund=knot_press_abund,
                    abund_smooth=self.abund_smooth,
                    pressure_grid=self.pressure_grid,
                    plotting=False,
                    contribution=True,
                    tau_cloud=tau_cloud,
                    cloud_wavel=self.cloud_wavel,
                    return_opacities=True,
                    eq_chem=self.eq_chem,
                )

                emission_contr = extra_out["emission_contribution"]

        elif chemistry == "equilibrium":
            # Calculate the petitRADTRANS spectrum for a clear atmosphere

            wavelength, flux, extra_out = calc_spectrum_clear(
                self.rt_object,
                self.pressure,
                temp,
                model_param["logg"],
                model_param["c_o_ratio"],
                model_param["metallicity"],
                p_quench,
                None,
                pressure_grid=self.pressure_grid,
                knot_press_abund=knot_press_abund,
                abund_smooth=self.abund_smooth,
                contribution=True,
                return_opacities=True,
                eq_chem=self.eq_chem,
            )

            emission_contr = extra_out["emission_contribution"]

        elif chemistry == "free":
            log_x_abund = {}

            if abund_nodes is None:
                for line_item in self.rt_object.line_species:
                    log_x_abund[line_item] = model_param[line_item]

            else:
                for line_item in self.rt_object.line_species:
                    for node_idx in range(abund_nodes):
                        log_x_abund[f"{line_item}_{node_idx}"] = model_param[
                            f"{line_item}_{node_idx}"
                        ]

            wavelength, flux, extra_out = calc_spectrum_clear(
                self.rt_object,
                self.pressure,
                temp,
                model_param["logg"],
                None,
                None,
                None,
                log_x_abund,
                knot_press_abund=knot_press_abund,
                abund_smooth=self.abund_smooth,
                pressure_grid=self.pressure_grid,
                contribution=True,
                return_opacities=True,
                eq_chem=self.eq_chem,
            )

            emission_contr = extra_out["emission_contribution"]

        if "radius" in model_param:
            # Calculate the planet mass from log(g) and radius

            model_param["mass"] = logg_to_mass(
                model_param["logg"], model_param["radius"]
            )

            # Scale the flux to the observer

            if "parallax" in model_param:
                scaling = (model_param["radius"] * constants.R_JUP) ** 2 / (
                    1e3 * constants.PARSEC / model_param["parallax"]
                ) ** 2

                flux *= scaling

            elif "distance" in model_param:
                scaling = (model_param["radius"] * constants.R_JUP) ** 2 / (
                    model_param["distance"] * constants.PARSEC
                ) ** 2

                flux *= scaling

        # Apply ISM extinction

        if "ism_ext" in model_param:
            if "ism_red" in model_param:
                ism_reddening = model_param["ism_red"]

            else:
                # Use default ISM reddening (R_V = 3.1) if ism_red is not provided
                ism_reddening = 3.1

            flux = apply_ism_ext(
                wavelength, flux, model_param["ism_ext"], ism_reddening
            )

        # Plot 2D emission contribution

        if plot_contribution:
            # Calculate the total optical depth (line and continuum opacities)
            # self.rt_object.calc_opt_depth(10.**model_param['logg'])

            # From Paul: The first axis of total_tau is the coordinate
            # of the cumulative opacity distribution function (ranging
            # from 0 to 1). A correct average is obtained by
            # multiplying the first axis with self.w_gauss, then
            # summing them. This is then the actual wavelength-mean.

            if self.scattering:
                # From petitRADTRANS: Only use 0 index for species
                # because for lbl or test_ck_shuffle_comp = True
                # everything has been moved into the 0th index
                w_gauss = self.rt_object._lines_loaded_opacities["weights_gauss"][
                    ..., np.newaxis, np.newaxis
                ]
                optical_depth = np.sum(
                    w_gauss * extra_out["optical_depths"][:, :, 0, :], axis=0
                )

            else:
                # TODO Is this correct?
                w_gauss = self.rt_object._lines_loaded_opacities["weights_gauss"][
                    ..., np.newaxis, np.newaxis, np.newaxis
                ]
                optical_depth = np.sum(
                    w_gauss * extra_out["optical_depths"][:, :, :, :], axis=0
                )

                # Sum over all species
                optical_depth = np.sum(optical_depth, axis=1)

            plt.rcParams["font.family"] = "serif"
            plt.rcParams["mathtext.fontset"] = "dejavuserif"

            plt.figure(figsize=(8.0, 4.0))
            gridsp = mpl.gridspec.GridSpec(1, 1)
            gridsp.update(wspace=0, hspace=0, left=0, right=1, bottom=0, top=1)

            ax = plt.subplot(gridsp[0, 0])

            ax.tick_params(
                axis="both",
                which="major",
                colors="black",
                labelcolor="black",
                direction="in",
                width=1,
                length=5,
                labelsize=12,
                top=True,
                bottom=True,
                left=True,
                right=True,
            )

            ax.tick_params(
                axis="both",
                which="minor",
                colors="black",
                labelcolor="black",
                direction="in",
                width=1,
                length=3,
                labelsize=12,
                top=True,
                bottom=True,
                left=True,
                right=True,
            )

            ax.set_xlabel("Wavelength (µm)", fontsize=13)
            ax.set_ylabel("Pressure (bar)", fontsize=13)

            ax.get_xaxis().set_label_coords(0.5, -0.09)
            ax.get_yaxis().set_label_coords(-0.07, 0.5)

            ax.set_yscale("log")

            ax.xaxis.set_major_locator(MultipleLocator(1.0))
            ax.xaxis.set_minor_locator(MultipleLocator(0.2))

            press_bar = 1e-6 * self.rt_object.pressures  # (Ba) -> (Bar)

            xx_grid, yy_grid = np.meshgrid(wavelength, press_bar)

            emission_contr = np.nan_to_num(emission_contr)

            ax.pcolormesh(
                xx_grid,
                yy_grid,
                emission_contr,
                cmap=plt.cm.bone_r,
                shading="gouraud",
            )

            photo_press = np.zeros(wavelength.shape[0])

            for i in range(photo_press.shape[0]):
                press_interp = interp1d(optical_depth[i, :], self.rt_object.pressures)
                photo_press[i] = press_interp(1.0) * 1e-6  # cgs to (bar)

            ax.plot(wavelength, photo_press, lw=0.5, color="gray")

            ax.set_xlim(np.amin(wavelength), np.amax(wavelength))
            ax.set_ylim(np.amax(press_bar), np.amin(press_bar))

            if isinstance(plot_contribution, str):
                plt.savefig(plot_contribution, bbox_inches="tight")
            else:
                plt.show()

            plt.clf()
            plt.close()

        # Convolve with a broadening kernel for vsin(i)

        if "vsini" in model_param:
            flux = rot_int_cmj(
                wavel=wavelength,
                flux=flux,
                vsini=model_param["vsini"],
                eps=0.0,
            )

        # Convolve the spectrum with a Gaussian LSF

        if spec_res is not None:
            flux = convolve_spectrum(wavelength, flux, spec_res)

        # Apply a radial velocity shift to the wavelengths

        if "rad_vel" in model_param:
            # Change speed of light from (m/s) to (km/s)
            wavelength *= 1.0 - model_param["rad_vel"] / (constants.LIGHT * 1e-3)

        # Resample the spectrum

        if wavel_resample is not None:
            flux = spectres_numba(
                wavel_resample,
                wavelength,
                flux,
                spec_errs=None,
                fill=np.nan,
                verbose=True,
            )

            wavelength = wavel_resample

        if hasattr(self.rt_object, "h_bol"):
            pressure = 1e-6 * self.rt_object.pressures  # (bar)
            f_bol = -4.0 * np.pi * self.rt_object.h_bol
            f_bol *= 1e-3  # (erg s-1 cm-2) -> (W m-2)
            bol_flux = np.column_stack((pressure, f_bol))

        else:
            bol_flux = None

        return create_box(
            boxtype="model",
            model="petitradtrans",
            wavelength=wavelength,
            flux=flux,
            parameters=model_param,
            quantity="flux",
            contribution=emission_contr,
            bol_flux=bol_flux,
            extra_out=extra_out,
        )

    @typechecked
    def get_flux(self, model_param: Dict[str, float]) -> Tuple[float, None]:
        """
        Function for calculating the filter-weighted flux density
        for the ``filter_name``.

        Parameters
        ----------
        model_param : dict
            Dictionary with the model parameters and values.

        Returns
        -------
        float
            Flux (W m-2 um-1).
        NoneType
            Uncertainty (W m-2 um-1). Always set to ``None``.
        """

        if "log_p_quench" in model_param:
            quenching = "pressure"
        else:
            quenching = None

        spectrum = self.get_model(model_param, quenching=quenching)

        synphot = SyntheticPhotometry(self.filter_name)

        return synphot.spectrum_to_flux(spectrum.wavelength, spectrum.flux)

    @typechecked
    def get_magnitude(self, model_param: Dict[str, float]) -> Tuple[float, None]:
        """
        Function for calculating the magnitude for the ``filter_name``.

        Parameters
        ----------
        model_param : dict
            Dictionary with the model parameters and values.

        Returns
        -------
        float
            Magnitude.
        NoneType
            Uncertainty. Always set to ``None``.
        """

        if "log_p_quench" in model_param:
            quenching = "pressure"
        else:
            quenching = None

        spectrum = self.get_model(model_param, quenching=quenching)

        synphot = SyntheticPhotometry(self.filter_name)
        app_mag, _ = synphot.spectrum_to_magnitude(spectrum.wavelength, spectrum.flux)

        return app_mag
